@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -52,6 +50,7 @@
 :- use_module(library(debug)).
 :- use_module(library(option)).
 :- use_module(library(arithmetic)).
+:- set_prolog_flag(generate_debug_info, false).
 
 /** <module> Setting management
 
