@@ -88,6 +88,10 @@ http_wrapper(Goal, In, Out, Close, Options) :-
 	    handler_with_output_to(Goal, Id, Request1, CGI, Error),
 	    cgi_close(CGI, State0, Error, Close)
 	;   Id = 0,
+	    (	debugging(http(request))
+	    ->	print_message(warning, ReqError)
+	    ;	true
+	    ),
 	    send_error(Out, State0, ReqError, Close),
 	    extend_request(Options, [], _)
 	).
@@ -196,7 +200,9 @@ send_error(Out, State0, Error, Close) :-
 	;   http_done(500,  E, 0, State0),
 	    throw(E)			% is that wise?
 	),
-	(   memberchk(connection(Close), HdrExtra)
+        (   Error = http_reply(switching_protocols(Goal, SwitchOptions))
+        ->  Close = switch_protocol(Goal, SwitchOptions)
+	;   memberchk(connection(Close), HdrExtra)
 	->  true
 	;   Close = close
 	).
