@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -31,6 +31,7 @@
 
 :- module(thread_httpd,
 	  [ http_current_server/2,	% ?:Goal, ?Port
+	    http_server_property/2,	% ?Port, ?Property
 	    http_server/2,		% :Goal, +Options
 	    http_workers/2,		% +Port, ?WorkerCount
 	    http_current_worker/2,	% ?Port, ?ThreadID
@@ -50,6 +51,19 @@
 :- use_module(http_stream).
 
 
+:- predicate_options(http_server/2, 2,
+		     [ port(integer),
+		       workers(positive_integer),
+		       timeout(number),
+		       keep_alive_timeout(number),
+		       pass_to(system:thread_create/3, 3)
+		     ]).
+:- predicate_options(http_spawn/2, 2,
+		     [ pool(atom),
+		       pass_to(system:thread_create/3, 3),
+		       pass_to(thread_pool:thread_create_in_pool/4, 4)
+		     ]).
+
 /** <module> Threaded HTTP server
 
 This library provides a multi-threaded Prolog-based HTTP server based on
@@ -64,13 +78,12 @@ for details.
 */
 
 :- meta_predicate
-	http_server(:, +),
-	http_current_server(:, ?),
-	http_spawn(:, +).
+	http_server(1, +),
+	http_current_server(1, ?),
+	http_spawn(0, +).
 
 :- dynamic
-	port_option/2,			% Port, Option
-	current_server/4,		% Port, Goal, Thread, Queue
+	current_server/5,		% Port, Goal, Thread, Queue, StartTime
 	queue_worker/2,			% Queue, ThreadID
 	queue_options/2.		% Queue, Options
 
@@ -78,26 +91,26 @@ for details.
 	make_socket_hook/3,
 	accept_hook/2,
 	close_hook/1,
-	open_client_hook/5.
+	open_client_hook/5,
+	http:create_pool/1.
 
 %%	http_server(:Goal, +Options) is det.
 %
 %	Create a server at Port that calls Goal for each parsed request.
 %	Options provide a list of options. Defined options are
 %
-%	| port(?Port)	     | - 	| Port to listen to |
-%	| workers(N)	     | 2 	| Define the number of worker threads |
-%	| timeout(S)	     | infinite	| Drop connections after inactivity   |
-%	| keep_alive_timeout | 10	| Drop Keep-Alive connection timeout  |
-%	| local(KBytes)	     | <CommandLine> |				    |
-%	| global(KBytes)     | <CommandLine> |				    |
+%	| port(?Port)	     | -	| Port to listen to		      |
+%	| workers(N)	     | 5	| Define the number of worker threads |
+%	| timeout(S)	     | 60	| Max inactivity for reading request  |
+%	| keep_alive_timeout | 2	| Drop Keep-Alive connection timeout  |
+%	| local(KBytes)	     | <CommandLine> |				      |
+%	| global(KBytes)     | <CommandLine> |				      |
 %	| trail(KBytes)      | <CommandLine> | Stack-sizes of worker threads  |
 
 http_server(Goal, Options) :-
 	strip_module(Goal, Module, G),
 	select_option(port(Port), Options, Options1), !,
 	make_socket(Port, Options1, Options2),
-	set_port_options(Port, Options2),
 	create_workers(Options2),
 	create_server(Module:G, Port, Options2).
 http_server(_Goal, _Options) :-
@@ -116,48 +129,72 @@ make_socket(Port, Options0, Options) :-
 	tcp_setopt(Socket, reuseaddr),
 	tcp_bind(Socket, Port),
 	tcp_listen(Socket, 5),
-	atom_concat('httpd@', Port, Queue),
+	make_addr_atom('httpd@', Port, Queue),
 	Options = [ queue(Queue),
 		    tcp_socket(Socket)
 		  | Options0
 		  ].
 
-create_server(Goal, Port, Options) :-
-	memberchk(queue(Queue), Options),
-	atom_concat('http@', Port, Alias),
-	thread_create(accept_server(Goal, Options), _,
-		      [ local(128),
-			global(128),
-			trail(128),
-			alias(Alias)
-		      ]),
-	assert(current_server(Port, Goal, Alias, Queue)).
-
-
-%%	set_port_options(+Port, +Options) is det.
+%%	make_addr_atom(+Prefix, +Address, -Atom) is det.
 %
-%	Register Options for the HTTP server at Port.
+%	Create an atom that identifies  the   server's  queue and thread
+%	resources.
 
-set_port_options(Port, Options) :-
-	retractall(port_option(Port, _)),
-	assert_port_options(Options, Port).
+make_addr_atom(Prefix, Address, Atom) :-
+	phrase(address_parts(Address), Parts),
+	atomic_list_concat([Prefix|Parts], Atom).
 
-assert_port_options([], _).
-assert_port_options([Name=Value|T], Port) :- !,
-	Opt =.. [Name,Value],
-	assert(port_option(Port, Opt)),
-	assert_port_options(T, Port).
-assert_port_options([Opt|T], Port) :- !,
-	assert(port_option(Port, Opt)),
-	assert_port_options(T, Port).
+address_parts(Atomic) -->
+	{ atomic(Atomic) }, !,
+	[Atomic].
+address_parts(Host:Port) --> !,
+	address_parts(Host), [:], address_parts(Port).
+address_parts(ip(A,B,C,D)) --> !,
+	[ A, '.', B, '.', C, '.', D ].
+
+%%	create_server(:Goal, +Address, +Options) is det.
+%
+%	Create the main server thread that runs accept_server/2 to
+%	listen to new requests.
+
+create_server(Goal, Port, Options) :-
+	get_time(StartTime),
+	memberchk(queue(Queue), Options),
+	make_addr_atom('http@', Port, Alias),
+	thread_create(accept_server(Goal, Options), _,
+		      [ alias(Alias)
+		      ]),
+	assert(current_server(Port, Goal, Alias, Queue, StartTime)).
 
 
 %%	http_current_server(:Goal, ?Port) is nondet.
 %
 %	True if Goal is the goal of a server at Port.
+%
+%	@deprecated Use http_server_property(Port, goal(Goal))
 
 http_current_server(Goal, Port) :-
-	current_server(Port, Goal, _, _).
+	current_server(Port, Goal, _, _, _).
+
+
+%%	http_server_property(?Port, ?Property) is nondet.
+%
+%	True if Property is a property of the HTTP server running at
+%	Port.  Defined properties are:
+%
+%	    * goal(:Goal)
+%	    Goal used to start the server. This is often
+%	    http_dispatch/1.
+%	    * start_time(?Time)
+%	    Time-stamp when the server was created.
+
+http_server_property(Port, Property) :-
+	server_property(Property, Port).
+
+server_property(goal(Goal), Port) :-
+	current_server(Port, Goal, _, _, _).
+server_property(start_time(Time), Port) :-
+	current_server(Port, _, _, _, Time).
 
 
 %%	http_workers(+Port, -Workers) is det.
@@ -169,7 +206,7 @@ http_current_server(Goal, Port) :-
 
 http_workers(Port, Workers) :-
 	must_be(integer, Port),
-	current_server(Port, _, _, Queue), !,
+	current_server(Port, _, _, Queue, _), !,
 	(   integer(Workers)
 	->  resize_pool(Queue, Workers)
 	;   findall(W, queue_worker(Queue, W), WorkerIDs),
@@ -187,7 +224,7 @@ http_workers(Port, _) :-
 %	statistics.
 
 http_current_worker(Port, ThreadID) :-
-	current_server(Port, _, _, Queue),
+	current_server(Port, _, _, Queue, _),
 	queue_worker(Queue, ThreadID).
 
 
@@ -199,7 +236,7 @@ http_current_worker(Port, ThreadID) :-
 accept_server(Goal, Options) :-
 	catch(accept_server2(Goal, Options), http_stop, true),
 	thread_self(Thread),
-	retract(current_server(_Port, _, Thread, _Queue)),
+	retract(current_server(_Port, _, Thread, _Queue, _StartTime)),
 	close_server_socket(Options).
 
 accept_server2(Goal, Options) :-
@@ -213,14 +250,17 @@ accept_server2(Goal, Options) :-
 	      ->  debug(http(connection), 'New HTTP connection from ~p', [Peer]),
 		  thread_send_message(Queue, tcp_client(Client, Goal, Peer)),
 		  fail
-	      ;   E == http_stop
-	      ->  throw(http_stop)
+	      ;   accept_rethrow_error(E)
+	      ->  throw(E)
 	      ;   print_message(error, E),
 		  fail
 	      )
 	  ;   print_message(error, goal_failed(tcp_accept(Socket, Client, Peer)))
 	  ).
 
+accept_rethrow_error(http_stop).
+accept_rethrow_error('$aborted').
+
 
 %%	close_server_socket(+Options)
 %
@@ -242,7 +282,7 @@ close_server_socket(Options) :-
 
 http_stop_server(Port, _Options) :-
 	http_workers(Port, 0),
-	current_server(Port, _, Thread, Queue),
+	current_server(Port, _, Thread, Queue, _Start),
 	retractall(queue_options(Queue, _)),
 	thread_signal(Thread, throw(http_stop)),
 	catch(connect(localhost:Port), _, true),
@@ -265,7 +305,7 @@ connect(Address) :-
 %	alias http_worker_N.
 
 create_workers(Options) :-
-	option(workers(N), Options, 2),
+	option(workers(N), Options, 5),
 	option(queue(Queue), Options),
 	catch(message_queue_create(Queue), _, true),
 	atom_concat(Queue, '_', AliasBase),
@@ -353,18 +393,24 @@ http_worker(Options) :-
 %	Opens the connection to the client in a worker from the message
 %	sent to the queue by accept_server/2.
 
-open_client(requeue(In, Out, Goal, ClOpts), _, Goal, In, Out, Opts, ClOpts) :- !,
+open_client(requeue(In, Out, Goal, ClOpts),
+	    _, Goal, In, Out, Opts, ClOpts) :- !,
 	memberchk(peer(Peer), ClOpts),
-	option(keep_alive_timeout(KeepAliveTMO), Opts, 5),
+	option(keep_alive_timeout(KeepAliveTMO), Opts, 2),
 	check_keep_alife_connection(In, KeepAliveTMO, Peer, In, Out).
-open_client(Message, Queue, Goal, In, Out, _Opts,
-	    [ pool(Queue, Goal, In, Out)
+open_client(Message, Queue, Goal, In, Out, Opts,
+	    [ pool(client(Queue, Goal, In, Out)),
+	      timeout(Timeout)
 	    | Options
 	    ]) :-
 	catch(open_client(Message, Goal, In, Out, Options),
 	      E, report_error(E)),
-	memberchk(peer(Peer), Options),
-	debug(http(connection), 'Opened connection from ~p', [Peer]).
+	option(timeout(Timeout), Opts, 60),
+	(   debugging(http(connection))
+	->  memberchk(peer(Peer), Options),
+	    debug(http(connection), 'Opened connection from ~p', [Peer])
+	;   true
+	).
 
 
 open_client(Message, Goal, In, Out, Options) :-
@@ -411,10 +457,30 @@ check_keep_alife_connection(In, TMO, Peer, In, Out) :-
 done_worker :-
 	thread_self(Self),
 	thread_property(Self, status(Status)),
-	retract(queue_worker(_Queue, Self)),
-	print_message(informational,
-		      httpd_stopped_worker(Self, Status)).
+	retract(queue_worker(Queue, Self)),
+	(   catch(recreate_worker(Status, Queue), _, fail)
+	->  thread_detach(Self),
+	    print_message(informational,
+			  httpd_restarted_worker(Self))
+	;   print_message(informational,
+			  httpd_stopped_worker(Self, Status))
+	).
 
+%%	recreate_worker(+Status, +Queue) is semidet.
+%
+%	Deal with the possibility that  threads are, during development,
+%	killed with abort/0. We  recreate  the   worker  to  avoid  that
+%	eventually we run out of workers.  If   we  are aborted due to a
+%	halt/0 call, thread_create/3 will raise a permission error.
+
+recreate_worker(exception(Error), Queue) :-
+	recreate_on_error(Error),
+	queue_options(Queue, Options),
+	atom_concat(Queue, '_', AliasBase),
+	create_workers(1, 1, Queue, AliasBase, Options).
+
+recreate_on_error('$aborted').
+recreate_on_error(time_limit_exceeded).
 
 %	thread_httpd:message_level(+Exception, -Level)
 %
@@ -444,7 +510,7 @@ current_message_level(Term, Level) :-
 
 http_requeue(Header) :-
 	requeue_header(Header, ClientOptions),
-	memberchk(pool(Queue, Goal, In, Out), ClientOptions),
+	memberchk(pool(client(Queue, Goal, In, Out)), ClientOptions),
 	thread_send_message(Queue, requeue(In, Out, Goal, ClientOptions)), !.
 http_requeue(Header) :-
 	debug(http(error), 'Re-queue failed: ~p', [Header]),
@@ -457,7 +523,7 @@ requeue_header([H|T0], [H|T]) :-
 requeue_header([_|T0], T) :-
 	requeue_header(T0, T).
 
-requeue_keep(pool(_,_,_,_)).
+requeue_keep(pool(_)).
 requeue_keep(peer(_)).
 requeue_keep(protocol(_)).
 
@@ -469,12 +535,23 @@ requeue_keep(protocol(_)).
 http_process(Goal, In, Out, Options) :-
 	debug(http(server), 'Running server goal ~p on ~p -> ~p',
 	      [Goal, In, Out]),
+	option(timeout(TMO), Options, 60),
+	set_stream(In, timeout(TMO)),
+	set_stream(Out, timeout(TMO)),
 	http_wrapper(Goal, In, Out, Connection,
 		     [ request(Request)
 		     | Options
 		     ]),
 	next(Connection, Request).
 
+next(switch_protocol(SwitchGoal, _SwitchOptions), Request) :- !,
+        memberchk(pool(client(_Queue, _Goal, In, Out)), Request),
+        (   catch(call(SwitchGoal, In, Out), E,
+		  (   print_message(error, E),
+		      fail))
+	->  true
+	;   http_close_connection(Request)
+	).
 next(spawned(ThreadId), _) :- !,
 	debug(http(spawn), 'Handler spawned to thread ~w', [ThreadId]).
 next(Connection, Request) :-
@@ -489,7 +566,7 @@ next(_, Request) :-
 %	Close connection associated to Request.  See also http_requeue/1.
 
 http_close_connection(Request) :-
-	memberchk(pool(_Queue, _Goal, In, Out), Request),
+	memberchk(pool(client(_Queue, _Goal, In, Out)), Request),
 	memberchk(peer(Peer), Request),
 	close_connection(Peer, In, Out).
 
@@ -514,21 +591,30 @@ close_connection(Peer, In, Out) :-
 %	    * pool(+Pool)
 %	    Interfaces to library(thread_pool), starting the thread
 %	    on the given pool.
-%	    * backlog(+MaxBacklog)
-%	    Reply using a 503 (service unavailable) if too many requests
-%	    are waiting in this pool.
+%
+%	If a pool does not exist, this predicate calls the multifile
+%	hook http:create_pool/1 to create it. If this predicate succeeds
+%	the operation is retried.
 
 http_spawn(Goal, Options) :-
-	select_option(pool(Pool), Options, Options1), !,
-	select_option(backlog(BackLog), Options1, ThreadOptions, infinite),
-	check_backlog(BackLog, Pool),
+	select_option(pool(Pool), Options, ThreadOptions), !,
 	current_output(CGI),
-	thread_create_in_pool(Pool,
-			      wrap_spawned(CGI, Goal), Id,
-			      [ detached(true)
-			      | ThreadOptions
-			      ]),
-	http_spawned(Id).
+	catch(thread_create_in_pool(Pool,
+				    wrap_spawned(CGI, Goal), Id,
+				    [ detached(true)
+				    | ThreadOptions
+				    ]),
+	      Error,
+	      true),
+	(   var(Error)
+	->  http_spawned(Id)
+	;   Error = error(resource_error(threads_in_pool(_)), _)
+	->  throw(http_reply(busy))
+	;   Error = error(existence_error(thread_pool, Pool), _),
+	    create_pool(Pool)
+	->  http_spawn(Goal, Options)
+	;   throw(Error)
+	).
 http_spawn(Goal, Options) :-
 	current_output(CGI),
 	thread_create(wrap_spawned(CGI, Goal), Id,
@@ -542,17 +628,19 @@ wrap_spawned(CGI, Goal) :-
 	http_wrap_spawned(Goal, Request, Connection),
 	next(Connection, Request).
 
-%%	check_backlog(+MaxBackLog, +Pool)
+%%	create_pool(+Pool)
 %
-%	Check whether the backlog in the pool  has been exceeded. If so,
-%	reply as =busy=, which causes an HTTP 503 response.
+%	Lazy  creation  of  worker-pools  for   the  HTTP  server.  This
+%	predicate calls the hook http:create_pool/1.   If the hook fails
+%	it creates a default pool of size 10.   This should suffice most
+%	typical usecases.
+
+create_pool(Pool) :-
+	http:create_pool(Pool), !.
+create_pool(Pool) :-
+	print_message(informational, httpd(created_pool(Pool))),
+	thread_pool_create(Pool, 10, []).
 
-check_backlog(BackLog, Pool) :-
-	integer(BackLog),
-	thread_pool_property(Pool, backlog(Waiting)),
-	Waiting > BackLog, !,
-	throw(http_reply(busy)).
-check_backlog(_, _).
 
 
 		 /*******************************
@@ -564,16 +652,11 @@ check_backlog(_, _).
 
 prolog:message(httpd_stopped_worker(Self, Status)) -->
 	[ 'Stopped worker ~p: ~p'-[Self, Status] ].
-
-
-		 /*******************************
-		 *	      XREF		*
-		 *******************************/
-
-:- multifile
-	prolog:meta_goal/2.
-:- dynamic
-	prolog:meta_goal/2.
-
-prolog:meta_goal(http_server(G, _), [G+1]).
-prolog:meta_goal(http_current_server(G, _), [G+1]).
+prolog:message(httpd_restarted_worker(Self)) -->
+	[ 'Replaced aborted worker ~p'-[Self] ].
+prolog:message(httpd(created_pool(Pool))) -->
+	[ 'Created thread-pool ~p of size 10'-[Pool], nl,
+	  'Create this pool at startup-time or define the hook ', nl,
+	  'http:create_pool/1 to avoid this message and create a ', nl,
+	  'pool that fits the usage-profile.'
+	].
