@@ -1,4 +1,4 @@
-/*  $Id: wrtest.pl,v 1.4 2007/06/06 09:34:07 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog SGML/XML parser
 
@@ -68,7 +68,7 @@ ml_file(sgml).
 ml_file(html).
 
 %%	blocked(+File)
-%	
+%
 %	List of test-files that are blocked.  These are either negative
 %	tests or tests involving SDATA.
 
@@ -140,12 +140,12 @@ cat(File, Encoding) :-
 	open(File, read, In, [encoding(Encoding)]),
 	copy_stream_data(In, current_output),
 	close(In).
-	
+
 %	eq(M1, M2)
-%	
+%
 %	Test two terms for equivalence.  The following mismatches are
 %	allowed:
-%	
+%
 %		* Order of attributes
 %		* Layout in `element-only' content
 
@@ -171,15 +171,15 @@ eq(A1, A2) :-
 	(   B1 == B2
 	->  true
 	;   format(user_error,
-		   'ERROR: CDATA differs:~n\
-		   \t~p~n\
+		   'ERROR: CDATA differs:~n\c
+		   \t~p~n\c
 		   \t~p~n',
 		   [B1, B2])
 	).
 eq(X, Y) :-
 	format(user_error,
-	       'ERROR: Content differs:~n\
-	       \t~p~n\
+	       'ERROR: Content differs:~n\c
+	       \t~p~n\c
 	       \t~p~n',
 	       [X, Y]).
 
@@ -188,8 +188,8 @@ att_eq(A1, A2) :-			% ordering is unimportant
 	sort(A2, S), !.
 att_eq(A1, A2) :-
 	format(user_error,
-	       'ERROR: Attribute lists differ:~n\
-	       \t~p~n\
+	       'ERROR: Attribute lists differ:~n\c
+	       \t~p~n\c
 	       \t~p~n',
 	       [A1, A2]).
 
