@@ -130,21 +130,31 @@ dlist(Codes, Tail, Codes, Tail).
 
 user:message_hook(Term, Level, Lines) :-
 	current_prolog_flag(message_ide, true),
-	ide_message(Term, Level, Lines).
+	ide_message(Term, Level, Lines),
+	fail.
+
+:- meta_predicate
+	pce(0).
 
 ide_message(Term, Level, Lines) :-
-	accept_level(Level),
+	accept_level(Level), !,
+	pce(pce_message(Term, Lines)).
+ide_message(make(reload(_Files)), _, _) :-
+	pce(clear_message_list).
+ide_message(emacs(consult(_File)), _, _) :-
+	pce(clear_message_list).
+
+pce_message(Term, Lines) :-
 	\+ object(@loading_emacs),
 	message_to_pce(Term, Lines, Location, String),
-	in_pce_thread(ide_message(Location, String)),
-	fail.
-ide_message(make(reload(_Files)), _, _) :-
-	in_pce_thread(clear_message_list),
-	fail.
-ide_message(emacs(consult(_File)), _, _) :-
-	in_pce_thread(clear_message_list),
-	fail.
+	ide_message(Location, String).
 
 accept_level(warning).
 accept_level(error).
 
+pce(Goal) :-
+	pce_thread(PceThread),
+	thread_self(PceThread), !,
+	Goal.
+pce(Goal) :-
+	in_pce_thread(Goal).
