@@ -104,6 +104,7 @@ class_variable(prompt_style, {mini_window,dialog}, dialog, "How to prompt").
 initialise(F, For:'emacs_buffer|emacs_view') :->
 	"Create window for buffer"::
 	send(F, send_super, initialise, 'PceEmacs', application := @emacs),
+	send(F, icon, image(resource(mode_x_icon))),
 	send(F, done_message, message(F, quit)),
 	send(F, append, new(MBD, emacs_mode_dialog)),
 
@@ -136,11 +137,24 @@ quit(F) :->
 	get(TW?members, size, Count),
 	(   Count == 1
 	->  send(F, destroy)
-	;   send(@display, confirm, 'Close %d tabs?', Count)
+	;   send(F, confirm, 'Close %d tabs?', Count)
 	->  send(F, destroy)
 	;   true
 	).
 
+confirm(F, Format:char_array, Args:any...) :->
+	"Confirm centered"::
+	new(D, dialog('Confirm action')),
+	String =.. [string, Format | Args ],
+	send(D, append, label(message, String)),
+	send(D, append, button(ok, message(D, return, ok))),
+	send(D, append, button(cancel, message(D, return, cancel))),
+	send(D, transient_for, F),
+	send(D, modal, transient),
+	get(D, confirm_centered, F?area?center, Rval),
+	send(D, destroy),
+	Rval == ok.
+
 editor_event(F, Ev:event) :->
 	"Delegate to the mini-window"::
 	get(F, member, mini_window, MW),
@@ -992,6 +1006,27 @@ label(E, Label:name) :->
 	get(E, device, View),
 	send(View, label, Label).
 
+size(E, Size:size) :->
+	"Set size, resizing the frame"::
+	get(E, frame, Frame),
+	get(Frame, size, size(FW,FH)),
+	get(E?area, size, size(W0,H0)),
+	send_super(E, size, Size),
+	get(E?area, size, size(W1,H1)),
+	FW1 is FW+W1-W0,
+	FH1 is FH+H1-H0,
+	send(Frame, size, size(FW1, FH1)).
+
+font(E, Font:font) :->
+	"Set font, resizing the frame"::
+	get(E, frame, Frame),
+	get(Frame, size, size(FW,FH)),
+	get(E?area, size, size(W0,H0)),
+	send_super(E, font, Font),
+	get(E?area, size, size(W1,H1)),
+	FW1 is FW+W1-W0,
+	FH1 is FH+H1-H0,
+	send(Frame, size, size(FW1, FH1)).
 
 looking_at(E, Re:regex, Where:[int], End:[int]) :->
 	"Test if regex macthes from the caret"::
