@@ -210,8 +210,16 @@ goto(FB, File:'file|node', Line:int) :->
 popup(FB, Id:any, Popup:popup) :<-
 	"Return popup from current node"::
 	get(FB, node, Id, Node),
-	send(Node, has_get_method, popup),
-	get(Node, popup, Popup).
+	(   send(Node, has_get_method, popup),
+	    get(Node, popup, Popup)
+	->  true
+	;   send(Node, instance_of, toc_directory)
+	->  new(Popup, popup(options)),
+	    send(Popup, append,
+		 menu_item(expand_all,
+			   message(Node, expand_all)))
+	).
+
 
 :- pce_group(edit).
 
@@ -262,10 +270,7 @@ event(FB, Ev:event) :->
 initialise(TF, File:file) :->
 	get(File, name, FileName),
 	canonical_source_file(FileName, Path),
-	(   source_file(Path)
-	->  Img = 'plloadedfile.xpm'
-	;   Img = 'plfile.xpm'
-	),
+	file_image(Path, Img),
 	file_base_name(FileName, Base),
 	send_super(TF, initialise, Base, Path, Img),
 	send(TF, name, Base).
@@ -277,6 +282,20 @@ loaded(TF) :->
 	get(TF, identifier, Path),
 	source_file(Path).
 
+included(TF) :->
+	get(TF, identifier, Path),
+	included_file(Path).
+
+file_image(Path, 'plloadedfile.xpm') :-
+	source_file(Path), !.
+file_image(Path, 'plincludedfile.xpm') :-
+	included_file(Path), !.
+file_image(_, 'plfile.xpm').
+
+included_file(Path) :-
+	source_file_property(Path, included_in(_,_)).
+
+
 module(TF, Module:name) :<-
 	"Return module defined in this file"::
 	get(TF, identifier, Path),
@@ -318,6 +337,9 @@ expand(TF) :->
 	;   true
 	).
 
+expand_all(_TF) :->
+	true.
+
 split_head(M:Head, Name, Arity, M) :- !,
 	callable(Head),
 	functor(Head, Name, Arity).
@@ -361,6 +383,8 @@ identify(TF) :->
 	get(TF, identifier, Path),
 	(   send(TF, loaded)
 	->  send(TF, report, status, 'Loaded file %s', Path)
+	;   send(TF, included)
+	->  send(TF, report, status, 'Included file %s', Path)
 	;   send(TF, report, status, 'File %s', Path)
 	).
 
@@ -932,11 +956,17 @@ image(predicate,	dcg,		'grammar.xpm').
 :- multifile
 	user:message_hook/3.
 
-image_of_load_state(start,	'loading.xpm').
-image_of_load_state(true,	'plloadedfile.xpm').
-image_of_load_state(false,	'loadfailed.xpm').
+image_of_load_state(start, _,	    'loading.xpm').
+image_of_load_state(true,  load,    'plloadedfile.xpm').
+image_of_load_state(true,  include, 'plincludedfile.xpm').
+image_of_load_state(false, _,	    'loadfailed.xpm').
 
 user:message_hook(load_file(What), _Kind, _Lines) :-
+	loading(What, load).
+user:message_hook(include_file(What), _Kind, _Lines) :-
+	loading(What, include).
+
+loading(What, How) :-
 	load_info(What, File, Stage),
 	prolog_overview_window(Win),
 	(   file_name_extension(_, qlf, File)
@@ -947,11 +977,14 @@ user:message_hook(load_file(What), _Kind, _Lines) :-
 	;   TheFile = File
 	),
 	get(Win, file_node, TheFile, Node),
-	image_of_load_state(Stage, Img),
-	send(Node, image, Img),
-	send(Node, flush),
+	image_of_load_state(Stage, How, Img),
+	in_pce_thread(update_image(Node, Img)),
 	fail.
 
+update_image(Node, Img) :-
+	send(Node, image, Img),
+	send(Node, flush).
+
 load_info(start(_Level, file(_, Path)),
 	  Path, start).
 load_info(failed(Spec),
@@ -963,4 +996,6 @@ load_info(failed(Spec),
 			   Path).
 load_info(done(_Level, file(_, Path), _, _, _, _),
 	  Path, true).
+load_info(done(_Level, file(_, Path)),
+	  Path, true).
 
