@@ -1,4 +1,4 @@
-/*  $Id: file.c,v 1.76 2007/01/23 19:22:18 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -64,8 +64,13 @@ initialiseFile(FileObj f, Name name, Name encoding)
   assign(f, status, NAME_closed);
   assign(f, bom, DEFAULT);
   assign(f, path, DEFAULT);
+#ifdef __WINDOWS__
+  assign(f, newline_mode, NAME_dos);
+#else
+  assign(f, newline_mode, NAME_posix);
+#endif
   f->fd = NULL;
-  
+
   kindFile(f, encoding);
 
   if ( isDefault(name) )
@@ -139,14 +144,14 @@ expandFileName(Name in)
 #if O_XOS
      wchar_t lng[MAXPATHLEN];
      char buf[MAXPATHLEN];
-  
+
      if ( _xos_long_file_nameW(expanded, lng, MAXPATHLEN) &&
 	  _xos_canonical_filenameW(lng, buf, sizeof(buf), 0) )
      { return UTF8ToName(buf);
      } else
      { errorPce(in, NAME_representation, NAME_nameTooLong);
        fail;
-     }    
+     }
 #else
      return WCToName(expanded, len);
 #endif
@@ -173,7 +178,7 @@ kindFile(FileObj f, Name encoding)
   { assign(f, encoding, encoding);
     assign(f, kind, NAME_text);
   }
-  
+
   succeed;
 }
 
@@ -211,7 +216,14 @@ loadFile(FileObj f, IOSTREAM *fd, ClassDef def)
     assign(f, encoding, (f->kind == NAME_binary ? NAME_octet : NAME_text));
   if ( !isDefault(f->bom) && !isBoolean(f->bom) )
     assign(f, bom, DEFAULT);
-
+  if ( !isName(f->newline_mode) )
+  {
+#ifdef __WINDOWS__
+    assign(f, newline_mode, NAME_dos);
+#else
+    assign(f, newline_mode, NAME_posix);
+#endif
+  }
   assign(f, status, NAME_closed);
   f->fd = NULL;
 
@@ -242,7 +254,7 @@ closeFile(FileObj f)
 
 
 status
-existsFile(FileObj f, Bool mustbefile)
+existsFile(FileObj f, BoolObj mustbefile)
 { STAT_TYPE buf;
   const char *fn = charArrayToFN((CharArray)f->name);
 
@@ -353,7 +365,7 @@ open_file(FileObj f, int access, ...)
 { va_list args;
   int mode;
   int fd = -1;
-  
+
   va_start(args, access);
   mode = va_arg(args, int);
   va_end(args);
@@ -375,7 +387,7 @@ copyFile(FileObj to, FileObj from)
   char buf[CPBUFSIZE];
   status rval;
   int n;
-  
+
   if ( (fdfrom = open_file(from, O_RDONLY|O_BINARY)) < 0 )
     fail;
   if ( (fdto = open_file(to, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, 0666)) < 0 )
@@ -492,11 +504,13 @@ accessFile(FileObj f, Name mode)
       m = R_OK;
     else if ( mode == NAME_write || mode == NAME_append )
       m = W_OK;
-#ifdef X_OK
     else /*if ( mode == NAME_execute )*/
+#ifdef X_OK
       m = X_OK;
+#else
+      m = R_OK;
 #endif
-  
+
     if ( access(strName(name), m) == 0 )
       succeed;
   }
@@ -547,13 +561,13 @@ doBOMFile(FileObj f)
     { if ( f->bom != OFF )
       { if ( ScheckBOM(f->fd) < 0 )
 	{ error:
-  
+
 	  reportErrorFile(f);
 	  closeFile(f);
-  
+
 	  fail;
 	}
-  
+
 	assign(f, bom, (f->fd->flags & SIO_BOM) ? ON : OFF);
 	if ( f->bom == ON )
 	  assign(f, encoding, encoding_to_name(f->fd->encoding));
@@ -576,7 +590,7 @@ openFile(FileObj f, Name mode, Name filter, CharArray extension)
 { CharArray path;
   Name name = getOsNameFile(f);
   char fdmode[3];
-  
+
   if ( f->status == NAME_tmpWrite )
   { if ( mode == NAME_write || mode == NAME_append )
     { assign(f, status, NAME_write);
@@ -686,7 +700,7 @@ removeFile(FileObj f)
     succeed;
   if ( existsFile(f, OFF) )
     return errorPce(f, NAME_removeFile, getOsErrorPce(PCE));
-  
+
   fail;
 }
 
@@ -770,14 +784,14 @@ append_file(FileObj f, String str)
   if ( f->encoding == NAME_octet )
   { if ( Sfwrite(str->s_text,
 		 isstrA(str) ? sizeof(charA) : sizeof(charW),
-		 str->size, 
+		 str->size,
 		 f->fd) != str->size )
       return reportErrorFile(f);
   } else
   { if ( isstrA(str) )
     { const charA *s = str->s_textA;
       const charA *e = &s[str->size];
-  
+
       for(; s<e; s++)
       { if ( Sputcode(*s, f->fd) < 0 )
 	  return reportErrorFile(f);
@@ -785,13 +799,13 @@ append_file(FileObj f, String str)
     } else
     { const charW *s = str->s_textW;
       const charW *e = &s[str->size];
-  
+
       for(; s<e; s++)
       { if ( Sputcode(*s, f->fd) < 0 )
 	  return reportErrorFile(f);
       }
     }
-  }  
+  }
 
   succeed;
 }
@@ -839,7 +853,7 @@ only slow ...
 
 static int
 statFile(FileObj f, STAT_TYPE *buf)
-{ 
+{
 #ifndef __WINDOWS__
   int fno;
 
@@ -879,7 +893,7 @@ getTimeFile(FileObj f, Name which)
   { errorPce(f, NAME_cannotStat, getOsErrorPce(PCE));
     fail;
   }
-  
+
   if ( which == NAME_modified )
     answer(CtoDate(buf.st_mtime));
   else
@@ -927,7 +941,7 @@ getReadLineFile(FileObj f)
     if ( c == '\n' )
       break;
   }
-      
+
   rval = StringToString(&tmp.s);
   str_tmp_done(&tmp);
 
@@ -997,7 +1011,7 @@ getCharacterFile(FileObj f)
     fail;
 
   chr = Sgetcode(f->fd);
-  
+
   answer(toInt(chr));
 }
 
@@ -1132,7 +1146,7 @@ storeStringFile(FileObj f, String s)
 
     DEBUG(NAME_save, Cprintf("Saved wide string, %ld chars\n", s->size));
   }
-  
+
   return checkErrorFile(f);
 }
 
@@ -1160,7 +1174,7 @@ waccess(const wchar_t *name, int m)
 
   str_set_n_wchar(&s, wcslen(name), (wchar_t *)name);
   ufn = stringToFN(&s);
-  
+
   DEBUG(NAME_find, Cprintf("find: trying \"%s\"\n", ufn));
 
   return access(ufn, m);
@@ -1185,7 +1199,7 @@ findFile(FileObj f, CharArray path, Name mode)
   base = charArrayToWC((CharArray)f->name, &bl);
   if ( base[0] == '.' )
     succeed;
-  
+
   if ( isDefault(mode) || mode == NAME_read )
     m = R_OK;
   else if ( mode == NAME_write || mode == NAME_append )
@@ -1283,6 +1297,8 @@ static vardecl var_file[] =
      NAME_filter, "Name of input/output filter used"),
   IV(NAME_bom, "[bool]", IV_BOTH,
      NAME_encoding, "Byte Order Mark"),
+  IV(NAME_newlineMode, "{posix,dos,detect}", IV_BOTH,
+     NAME_encoding, "Newline representation"),
   IV(NAME_fd, "alien:FILE *", IV_NONE,
      NAME_internal, "Unix file (stream) handle")
 };
@@ -1365,7 +1381,7 @@ static getdecl get_file[] =
 #define rc_file NULL
 /*
 static classvardecl rc_file[] =
-{ 
+{
 };
 */
 
