@@ -626,8 +626,10 @@ indirect_rlc_hwnd()
 	{ }
 #define PROLOG_INSTALL_REDRAW_FUNCTION(f) \
 	{ old_update_hook = indirect_rlc_update_hook(f); }
+#ifndef O_SHAREDLIBRARY
 #define O_SHAREDLIBRARY
 #endif
+#endif
 
 #ifdef O_SHAREDLIBRARY
 #ifndef PROLOG_ITF_INIT
@@ -2599,10 +2601,10 @@ PrologCall(PceGoal goal)
 	goto error;
 
       if ( goal->flags & PCE_GF_SEND )
-	rval = PL_call_predicate(MODULE_user, DebugMode,
+	rval = PL_call_predicate(MODULE_user, DebugMode|PL_Q_PASS_EXCEPTION,
 				 PREDICATE_send_implementation, av);
       else
-      { rval = PL_call_predicate(MODULE_user, DebugMode,
+      { rval = PL_call_predicate(MODULE_user, DebugMode|PL_Q_PASS_EXCEPTION,
 				 PREDICATE_get_implementation, av);
 	if ( rval )
 	{ if ( !get_answer_object(goal, av+3, goal->return_type, &goal->rval) )
@@ -2791,6 +2793,7 @@ better solution for signal handling is to be searched for (also avoiding
 the possibility of reentrance at moments this is not allowed in PCE ...
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#ifndef __WINDOWS__
 #ifndef PROLOG_DISPATCH_INPUT
 #define PROLOG_DISPATCH_INPUT 1
 #define PROLOG_DISPATCH_TIMEOUT 0
@@ -2807,6 +2810,7 @@ pce_dispatch(int fd)
 
   return PROLOG_DISPATCH_TIMEOUT;
 }
+#endif /*__WINDOWS__*/
 
 #ifdef SICSTUS
 
@@ -3273,6 +3277,7 @@ detach_thread(void *closure)
 { pceMTdetach();
 }
 
+
 foreign_t
 pl_pce_init(term_t a)
 { char **argv;
