@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
@@ -48,6 +46,7 @@ in this array.
 #include <float.h>
 #ifdef _MSC_VER
 #define isnan(x) _isnan(x)
+#define copysign(x,y) _copysign(x,y)
 #endif
 #endif
 #ifdef HAVE_IEEEFP_H
@@ -58,6 +57,10 @@ in this array.
 #define HAVE_FPCLASSIFY 1
 #endif
 
+#ifdef __WINDOWS__
+#include <wincrypt.h>
+#endif
+
 #undef LD
 #define LD LOCAL_LD
 
@@ -1030,6 +1033,8 @@ promoteIntNumber(Number n)
 		*     ARITHMETIC FUNCTIONS      *
 		*********************************/
 
+static int ar_u_minus(Number n1, Number r);
+
 int
 ar_add_ui(Number n, intptr_t add)
 { switch(n->type)
@@ -1139,8 +1144,8 @@ ar_minus(Number n1, Number n2, Number r)
   { case V_INTEGER:
     { r->value.i = n1->value.i - n2->value.i;
 
-      if ( (n1->value.i > 0 && n2->value.i < 0 && r->value.i <= 0) ||
-	   (n1->value.i < 0 && n2->value.i > 0 && r->value.i >= 0) )
+      if ( (n1->value.i >= 0 && n2->value.i < 0 && r->value.i <= 0) ||
+	   (n1->value.i < 0  && n2->value.i > 0 && r->value.i >= 0) )
       {					/* overflow */
 	if ( !promoteIntNumber(n1) ||
 	     !promoteIntNumber(n2) )
@@ -1343,7 +1348,7 @@ ar_shift(Number n1, Number n2, Number r, int dir)
 	}
       } else
       { if ( shift >= (long)sizeof(int64_t)*8 )
-	  r->value.i = (r->value.i >= 0 ? 0 : -1);
+	  r->value.i = (n1->value.i >= 0 ? 0 : -1);
 	else
 	  r->value.i = n1->value.i >> shift;
       }
@@ -1820,6 +1825,49 @@ ar_div(Number n1, Number n2, Number r)
 }
 
 
+#ifndef HAVE_SIGNBIT				/* differs for -0.0 */
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+signbit() and copysign() are part of C99.   These  should be provided by
+most C compilers, but Microsoft decided  not   to  adopt  C99 (it is now
+2012).
+
+Note that there is no autoconf support  to verify that floats conform to
+the IEE754 representation,  but  they  typically   do  these  days.  See
+http://www.gnu.org/software/autoconf/manual/autoconf-2.67/html_node/Floating-Point-Portability.html
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+#define IEEE754 1
+
+#ifdef IEEE754
+static inline int
+signbit(double f)
+{ union
+  { double f;
+    int64_t i;
+  } v;
+
+  v.f = f;
+  return v.i < 0;
+}
+
+#ifndef copysign
+double
+copysign(double x, double y)
+{ union { double f; uint64_t i; } ux, uy;
+  const uint64_t smask = (uint64_t)1<<(sizeof(uint64_t)*8-1);
+
+  ux.f = x;
+  uy.f = y;
+  ux.i &= ~smask;
+  ux.i |= (uy.i & smask);
+
+  return ux.f;
+}
+#endif
+#else
+#error "Don't know how to support signbit() and copysign()"
+#endif
+#endif
+
 int
 ar_sign_i(Number n1)
 { switch(n1->type)
@@ -1831,8 +1879,6 @@ ar_sign_i(Number n1)
     case V_MPQ:
       return mpq_sgn(n1->value.mpq);
 #endif
-    case V_FLOAT:
-      return (n1->value.f < 0.0 ? -1 : n1->value.f > 0.0 ? 1 : 0);
     default:
       assert(0);
       fail;
@@ -1841,12 +1887,59 @@ ar_sign_i(Number n1)
 
 static int
 ar_sign(Number n1, Number r)
-{ r->value.i = ar_sign_i(n1);
-  r->type = V_INTEGER;
+{ if ( n1->type == V_FLOAT )
+  { r->value.f = n1->value.f < 0 ? -1.0 : n1->value.f > 0.0 ? 1.0 : 0.0;
+    r->type = V_FLOAT;
+  } else
+  { r->value.i = ar_sign_i(n1);
+    r->type = V_INTEGER;
+  }
+
   succeed;
 }
 
 
+static int
+ar_signbit(Number n)
+{ switch(n->type)
+  { case V_INTEGER:
+      return n->value.i	< 0 ? -1 : 1;
+#ifdef O_GMP
+    case V_MPZ:
+    { int i = mpz_sgn(n->value.mpz);
+      return i < 0 ? -1 : 1;
+    }
+    case V_MPQ:
+    { int i = mpq_sgn(n->value.mpq);
+      return i < 0 ? -1 : 1;
+    }
+#endif
+    case V_FLOAT:
+      return signbit(n->value.f) ? -1 : 1;
+    default:
+      assert(0);
+      return 0;
+  }
+}
+
+
+static int
+ar_copysign(Number n1, Number n2, Number r)
+{
+  if ( n1->type == V_FLOAT && n2->type == V_FLOAT )
+  { r->value.f = copysign(n1->value.f, n2->value.f);
+    r->type = V_FLOAT;
+  } else
+  { if ( ar_signbit(n1) != ar_signbit(n2) )
+      return ar_u_minus(n1, r);
+    else
+      cpNumber(r, n1);
+  }
+
+  return TRUE;
+}
+
+
 static int
 ar_rem(Number n1, Number n2, Number r)
 { if ( !toIntegerNumber(n1, 0) )
@@ -2477,7 +2570,10 @@ ar_abs(Number n1, Number r)
       break;
 #endif
     case V_FLOAT:
-    { r->value.f = abs(n1->value.f);
+    { if ( signbit(n1->value.f) )
+	r->value.f = -n1->value.f;
+      else
+	r->value.f = n1->value.f;
       r->type = V_FLOAT;
       break;
     }
@@ -2782,7 +2878,7 @@ ar_truncate(Number n1, Number r)
 static int
 seed_from_dev(const char *dev ARG_LD)
 { int done = FALSE;
-#ifdef S_ISCHR
+#if defined(S_ISCHR) && !defined(__WINDOWS__)
   int fd;
 
   if ( (fd=open(dev, O_RDONLY)) )
@@ -2828,7 +2924,7 @@ seed_from_dev(const char *dev ARG_LD)
 static int
 seed_from_crypt_context(ARG1_LD)
 {
-#ifdef _MSC_VER
+#ifdef __WINDOWS__
   HCRYPTPROV hCryptProv;
   char *user_name = "seed_random";
   BYTE seedarray[RAND_SEED_LEN];
@@ -2896,7 +2992,12 @@ init_random(ARG1_LD)
 #ifdef O_GMP
   if ( !LD->arith.random.initialised )
   { LD->gmp.persistent++;
+#ifdef HAVE_GMP_RANDINIT_MT
+#define O_RANDOM_STATE 1
     gmp_randinit_mt(LD->arith.random.state);
+#else
+    gmp_randinit_default(LD->arith.random.state);
+#endif
     LD->arith.random.initialised = TRUE;
     seed_random(PASS_LD1);
     LD->gmp.persistent--;
@@ -2928,7 +3029,7 @@ PRED_IMPL("set_random", 1, set_random, 0)
 
 	if ( !PL_get_number(arg, &n) )
 	  return PL_error(NULL, 0, "integer or 'random'",
-			  ERR_TYPE, ATOM_seed, a);
+			  ERR_TYPE, ATOM_seed, arg);
 	switch(n.type)
 	{
 #ifdef O_GMP
@@ -2947,10 +3048,10 @@ PRED_IMPL("set_random", 1, set_random, 0)
 	  }
 #endif
 	  default:
-	    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_seed, a);
+	    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_seed, arg);
 	}
       }
-#ifdef O_GMP
+#ifdef O_RANDOM_STATE
     } else if ( name == ATOM_state )
     { number n;
 
@@ -2972,7 +3073,7 @@ PRED_IMPL("set_random", 1, set_random, 0)
 }
 
 
-#ifdef O_GMP
+#ifdef O_RANDOM_STATE
 static
 PRED_IMPL("random_property", 1, random_property, 0)
 { PRED_LD
@@ -3055,16 +3156,18 @@ ar_random_float(Number r)
 
   init_random(PASS_LD1);
 
+  do
+  {
 #ifdef O_GMP
-{ mpf_t rop;
-  mpf_init2(rop, sizeof(double)*8);
-  mpf_urandomb(rop, LD->arith.random.state, sizeof(double)*8);
-  r->value.f = mpf_get_d(rop);
-  mpf_clear(rop);
-}
+    mpf_t rop;
+    mpf_init2(rop, sizeof(double)*8);
+    mpf_urandomb(rop, LD->arith.random.state, sizeof(double)*8);
+    r->value.f = mpf_get_d(rop);
+    mpf_clear(rop);
 #else
   r->value.f = _PL_Random()/(float)UINT64_MAX;
 #endif
+  } while (r->value.f == 0.0);
 
   r->type = V_FLOAT;
   succeed;
@@ -3182,78 +3285,80 @@ typedef struct
   ArithF	function;
 } ar_funcdef;
 
-#define ADD(functor, func) { functor, func }
+#define F_ISO 0x1
+#define ADD(functor, func, flags) { functor, func }
 
 static const ar_funcdef ar_funcdefs[] = {
-  ADD(FUNCTOR_plus2,		pl_ar_add),
-  ADD(FUNCTOR_minus2,		ar_minus),
-  ADD(FUNCTOR_star2,		ar_mul),
-  ADD(FUNCTOR_divide2,		ar_divide),
+  ADD(FUNCTOR_plus2,		pl_ar_add, F_ISO),
+  ADD(FUNCTOR_minus2,		ar_minus, F_ISO),
+  ADD(FUNCTOR_star2,		ar_mul, F_ISO),
+  ADD(FUNCTOR_divide2,		ar_divide, F_ISO),
 #ifdef O_GMP
-  ADD(FUNCTOR_rational1,	ar_rational),
-  ADD(FUNCTOR_rationalize1,	ar_rationalize),
-  ADD(FUNCTOR_rdiv2,		ar_rdiv),
+  ADD(FUNCTOR_rational1,	ar_rational, 0),
+  ADD(FUNCTOR_rationalize1,	ar_rationalize, 0),
+  ADD(FUNCTOR_rdiv2,		ar_rdiv, 0),
 #endif
-  ADD(FUNCTOR_minus1,		ar_u_minus),
-  ADD(FUNCTOR_plus1,		ar_u_plus),
-  ADD(FUNCTOR_abs1,		ar_abs),
-  ADD(FUNCTOR_max2,		ar_max),
-  ADD(FUNCTOR_min2,		ar_min),
+  ADD(FUNCTOR_minus1,		ar_u_minus, F_ISO),
+  ADD(FUNCTOR_plus1,		ar_u_plus, F_ISO),
+  ADD(FUNCTOR_abs1,		ar_abs, F_ISO),
+  ADD(FUNCTOR_max2,		ar_max, F_ISO),
+  ADD(FUNCTOR_min2,		ar_min, F_ISO),
 
-  ADD(FUNCTOR_mod2,		ar_mod),
-  ADD(FUNCTOR_rem2,		ar_rem),
-  ADD(FUNCTOR_div2,		ar_div),
-  ADD(FUNCTOR_gdiv2,		ar_tdiv),
-  ADD(FUNCTOR_gcd2,		ar_gcd),
-  ADD(FUNCTOR_sign1,		ar_sign),
+  ADD(FUNCTOR_mod2,		ar_mod, F_ISO),
+  ADD(FUNCTOR_rem2,		ar_rem, F_ISO),
+  ADD(FUNCTOR_div2,		ar_div, F_ISO),
+  ADD(FUNCTOR_gdiv2,		ar_tdiv, 0),
+  ADD(FUNCTOR_gcd2,		ar_gcd, 0),
+  ADD(FUNCTOR_sign1,		ar_sign, F_ISO),
 
-  ADD(FUNCTOR_and2,		ar_conjunct),
-  ADD(FUNCTOR_bitor2,		ar_disjunct),
-  ADD(FUNCTOR_rshift2,		ar_shift_right),
-  ADD(FUNCTOR_lshift2,		ar_shift_left),
-  ADD(FUNCTOR_xor2,		ar_xor),
-  ADD(FUNCTOR_backslash1,	ar_negation),
+  ADD(FUNCTOR_and2,		ar_conjunct, F_ISO),
+  ADD(FUNCTOR_bitor2,		ar_disjunct, F_ISO),
+  ADD(FUNCTOR_rshift2,		ar_shift_right, F_ISO),
+  ADD(FUNCTOR_lshift2,		ar_shift_left, F_ISO),
+  ADD(FUNCTOR_xor2,		ar_xor, F_ISO),
+  ADD(FUNCTOR_backslash1,	ar_negation, F_ISO),
 
-  ADD(FUNCTOR_random1,		ar_random),
+  ADD(FUNCTOR_random1,		ar_random, 0),
 #ifdef O_GMP
-  ADD(FUNCTOR_random_float0,	ar_random_float),
+  ADD(FUNCTOR_random_float0,	ar_random_float, 0),
 #endif
 
-  ADD(FUNCTOR_integer1,		ar_integer),
-  ADD(FUNCTOR_round1,		ar_integer),
-  ADD(FUNCTOR_truncate1,	ar_truncate),
-  ADD(FUNCTOR_float1,		ar_float),
-  ADD(FUNCTOR_floor1,		ar_floor),
-  ADD(FUNCTOR_ceil1,		ar_ceil),
-  ADD(FUNCTOR_ceiling1,		ar_ceil),
-  ADD(FUNCTOR_float_fractional_part1, ar_float_fractional_part),
-  ADD(FUNCTOR_float_integer_part1, ar_float_integer_part),
+  ADD(FUNCTOR_integer1,		ar_integer, F_ISO),
+  ADD(FUNCTOR_round1,		ar_integer, F_ISO),
+  ADD(FUNCTOR_truncate1,	ar_truncate, F_ISO),
+  ADD(FUNCTOR_float1,		ar_float, F_ISO),
+  ADD(FUNCTOR_floor1,		ar_floor, F_ISO),
+  ADD(FUNCTOR_ceil1,		ar_ceil, F_ISO),
+  ADD(FUNCTOR_ceiling1,		ar_ceil, F_ISO),
+  ADD(FUNCTOR_float_fractional_part1, ar_float_fractional_part, F_ISO),
+  ADD(FUNCTOR_float_integer_part1, ar_float_integer_part, F_ISO),
+  ADD(FUNCTOR_copysign2,	ar_copysign, 0),
 
-  ADD(FUNCTOR_sqrt1,		ar_sqrt),
-  ADD(FUNCTOR_sin1,		ar_sin),
-  ADD(FUNCTOR_cos1,		ar_cos),
-  ADD(FUNCTOR_tan1,		ar_tan),
-  ADD(FUNCTOR_asin1,		ar_asin),
-  ADD(FUNCTOR_acos1,		ar_acos),
-  ADD(FUNCTOR_atan1,		ar_atan),
-  ADD(FUNCTOR_atan2,		ar_atan2),
-  ADD(FUNCTOR_atan22,		ar_atan2),
-  ADD(FUNCTOR_log1,		ar_log),
-  ADD(FUNCTOR_exp1,		ar_exp),
-  ADD(FUNCTOR_log101,		ar_log10),
-  ADD(FUNCTOR_hat2,		ar_pow),
-  ADD(FUNCTOR_doublestar2,	ar_pow),
-  ADD(FUNCTOR_pi0,		ar_pi),
-  ADD(FUNCTOR_e0,		ar_e),
-  ADD(FUNCTOR_epsilon0,		ar_epsilon),
+  ADD(FUNCTOR_sqrt1,		ar_sqrt, F_ISO),
+  ADD(FUNCTOR_sin1,		ar_sin, F_ISO),
+  ADD(FUNCTOR_cos1,		ar_cos, F_ISO),
+  ADD(FUNCTOR_tan1,		ar_tan, F_ISO),
+  ADD(FUNCTOR_asin1,		ar_asin, F_ISO),
+  ADD(FUNCTOR_acos1,		ar_acos, F_ISO),
+  ADD(FUNCTOR_atan1,		ar_atan, F_ISO),
+  ADD(FUNCTOR_atan2,		ar_atan2, 0),
+  ADD(FUNCTOR_atan22,		ar_atan2, F_ISO),
+  ADD(FUNCTOR_log1,		ar_log, F_ISO),
+  ADD(FUNCTOR_exp1,		ar_exp, F_ISO),
+  ADD(FUNCTOR_log101,		ar_log10, 0),
+  ADD(FUNCTOR_hat2,		ar_pow, F_ISO),
+  ADD(FUNCTOR_doublestar2,	ar_pow, F_ISO),
+  ADD(FUNCTOR_pi0,		ar_pi, F_ISO),
+  ADD(FUNCTOR_e0,		ar_e, 0),
+  ADD(FUNCTOR_epsilon0,		ar_epsilon, 0),
 
-  ADD(FUNCTOR_cputime0,		ar_cputime),
-  ADD(FUNCTOR_msb1,		ar_msb),
-  ADD(FUNCTOR_lsb1,		ar_lsb),
-  ADD(FUNCTOR_popcount1,	ar_popcount),
-  ADD(FUNCTOR_powm3,		ar_powm),
+  ADD(FUNCTOR_cputime0,		ar_cputime, 0),
+  ADD(FUNCTOR_msb1,		ar_msb, 0),
+  ADD(FUNCTOR_lsb1,		ar_lsb, 0),
+  ADD(FUNCTOR_popcount1,	ar_popcount, 0),
+  ADD(FUNCTOR_powm3,		ar_powm, 0),
 
-  ADD(FUNCTOR_eval1,		ar_eval)
+  ADD(FUNCTOR_eval1,		ar_eval, 0)
 };
 
 #undef ADD
@@ -3468,7 +3573,7 @@ BeginPredDefs(arith)
   PRED_DEF("plus", 3, plus, 0)
   PRED_DEF("between", 3, between, PL_FA_NONDETERMINISTIC)
   PRED_DEF("set_random", 1, set_random, 0)
-#ifdef O_GMP
+#ifdef O_RANDOM_STATE
   PRED_DEF("random_property", 1, random_property, 0)
 #endif
 EndPredDefs
