@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2010, University of Amsterdam, VU University Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam,
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,8 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+    MA  02110-1301  USA
 */
 
 #ifndef _PL_INCLUDE_H
@@ -54,7 +54,15 @@
 #else
 #define C_PLLIB	    "swipl.lib"
 #endif
-#else
+#else					/* !_MSC_VER  */
+#ifdef __WINDOWS__			/* I.e., MinGW */
+#define C_LIBS	     ""
+#define C_STATICLIBS ""
+#define C_CC	     "gcc"
+#define C_CFLAGS     ""
+#define C_PLLIB	     "-lswipl"
+#define C_LDFLAGS    ""
+#endif
 #include <parms.h>			/* pick from the working dir */
 #endif
 
@@ -89,6 +97,8 @@ handy for it someone wants to add a data type to the system.
       Include arithmetic compiler (compiles is/2, >/2, etc. into WAM).
   O_COMPILE_IS
       Compile Var = Value in the body.
+  O_CALL_AT_MODULE
+      Support the Goal@Module control-structure
   O_LABEL_ADDRESSES
       Means we can pick up the address of a label in  a function using
       the var  = `&&label' construct  and jump to  it using goto *var;
@@ -104,11 +114,10 @@ handy for it someone wants to add a data type to the system.
       Use `logical' update-view for dynamic predicates rather then the
       `immediate' update-view of older Prolog systems.
   O_PLMT
-      Include support for multi-threaded Prolog application.  Currently
-      very incomplete and only for the POSIX thread library.
+      Include support for multi-threading. Too much of the system relies
+      on this now, so it cannot be disabled without significant work.
   O_LARGEFILES
-      Supports files >2GB (if the OS provides it, currently requires
-      the GNU c library).
+      Supports files >2GB on 32-bit systems (if the OS provides it).
   O_ATTVAR
       Include support for attributes variables.
       This option requires O_DESTRUCTIVE_ASSIGNMENT.
@@ -125,6 +134,7 @@ handy for it someone wants to add a data type to the system.
 #define O_SOFTCUT		1
 #define O_COMPILE_ARITH		1
 #define O_COMPILE_IS		1
+#define O_CALL_AT_MODULE	1
 #define O_STRING		1
 #define O_CATCHTHROW		1
 #define O_DEBUGGER		1
@@ -224,7 +234,9 @@ gcc.
 
 /* AIX requires this to be the first thing in the file.  */
 #ifdef __GNUC__
-# define alloca __builtin_alloca
+# ifndef alloca
+#  define alloca __builtin_alloca
+# endif
 #else
 # if HAVE_ALLOCA_H
 #  include <alloca.h>
@@ -246,6 +258,9 @@ void *alloca ();
 #endif
 
 #include <sys/types.h>
+#if __MINGW32__
+typedef _sigset_t sigset_t;
+#endif
 #include <setjmp.h>
 #ifdef ASSERT_H_REQUIRES_STDIO_H
 #include <stdio.h>
@@ -503,14 +518,9 @@ Operator types.  NOTE: if you change OP_*, check operatorTypeToAtom()!
 #define OP_YFX	(0x70|OP_INFIX)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Type fields.  These codes are  included  in  a  number  of  the  runtime
-structures  at  a  fixed  point, so the runtime environment can tell the
-difference.
+Magic for assertions.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define HeapMagic(n)	((n) | 0x25678000)
-#define PROCEDURE_TYPE	HeapMagic(1)	/* a procedure */
-#define RECORD_TYPE	HeapMagic(2)	/* a record list */
 #define StackMagic(n)	((n) | 0x98765000)
 #define QID_MAGIC	StackMagic(1)	/* Query frame */
 
@@ -788,43 +798,40 @@ with one operation, it turns out to be faster as well.
 #define clear(s, a)		((s)->flags &= ~(a))
 #define clearFlags(s)		((s)->flags = 0)
 
-#define NONDETERMINISTIC	(0x00000001L) /* predicate */
-#define DISCONTIGUOUS		(0x00000002L) /* predicate */
-#define DYNAMIC			(0x00000004L) /* predicate */
-#define FOREIGN			(0x00000008L) /* predicate */
-#define HIDE_CHILDS		(0x00000010L) /* predicate */
-#define MULTIFILE		(0x00000020L) /* predicate */
-#define P_NOPROFILE		(0x00000040L) /* predicate */
-#define SPY_ME			(0x00000080L) /* predicate */
-#define SYSTEM			(0x00000100L) /* predicate, module */
-#define TRACE_ME		(0x00000200L) /* predicate */
-#define P_TRANSPARENT		(0x00000400L) /* predicate */
-#define GC_SAFE			(0x00000800L) /* predicate */
-#define TRACE_CALL		(0x00001000L) /* predicate */
-#define TRACE_REDO		(0x00002000L) /* predicate */
-#define TRACE_EXIT		(0x00004000L) /* predicate */
-#define TRACE_FAIL		(0x00008000L) /* predicate */
-					/* This may be changed later ... */
-#define LOCKED			(SYSTEM)      /* predicate */
-#define FILE_ASSIGNED		(0x00010000L) /* predicate */
-#define VOLATILE		(0x00020000L) /* predicate */
-#define AUTOINDEX		(0x00040000L) /* predicate */
-#define NEEDSCLAUSEGC		(0x00080000L) /* predicate */
-		      /* unused (0x00100000L) */
-#define P_VARARG		(0x00200000L) /* predicate */
-		      /* unused	(0x00400000L) */
-#define P_REDEFINED		(0x00800000L) /* predicate */
-#define PROC_DEFINED		(DYNAMIC|FOREIGN|MULTIFILE|DISCONTIGUOUS)
-#define P_THREAD_LOCAL		(0x01000000L) /* predicate */
-#define P_FOREIGN_CREF		(0x02000000L) /* predicate */
-#define P_DIRTYREG		(0x04000000L) /* predicate */
-#define P_ISO			(0x08000000L) /* predicate */
-#define P_META			(0x10000000L) /* predicate */
-#define P_MFCONTEXT		(0x20000000L) /* predicate */
-#define P_PUBLIC		(0x40000000L) /* predicate */
+/* Flags on predicates (packed in unsigned int */
 
-#define ERASED			(0x0001) /* clause, record */
-					 /* clause flags */
+#define P_FOREIGN		(0x00000020) /* Implemented in C */
+#define P_NONDET		(0x00000040) /* Foreign: nondet */
+#define P_VARARG		(0x00000080) /* Foreign: use alt calling API */
+#define P_FOREIGN_CREF		(0x00000100) /* Foreign: ndet ctx is clause */
+#define P_DYNAMIC		(0x00000200) /* Dynamic predicate */
+#define P_THREAD_LOCAL		(0x00000400) /* Thread local dynamic predicate */
+#define P_VOLATILE		(0x00000800) /* Clauses are not saved */
+#define P_DISCONTIGUOUS		(0x00001000) /* Clauses are not together */
+#define P_MULTIFILE		(0x00002000) /* Clauses are in multiple files */
+#define P_PUBLIC		(0x00004000) /* Called from somewhere */
+#define P_ISO			(0x00008000) /* Part of the ISO standard */
+#define P_LOCKED		(0x00010000) /* Locked as system predicate */
+#define P_NOPROFILE		(0x00020000) /* Profile children, not me */
+#define P_TRANSPARENT		(0x00040000) /* Inherit calling module */
+#define P_META			(0x00080000) /* Has meta_predicate declaration */
+#define P_MFCONTEXT		(0x00100000) /* Used for Goal@Module */
+#define P_DIRTYREG		(0x00200000) /* Part of GD->procedures.dirty */
+#define NEEDSCLAUSEGC		(0x00400000) /* Holds erased clauses */
+#define HIDE_CHILDS		(0x00800000) /* Hide children from tracer */
+#define SPY_ME			(0x01000000) /* Spy point placed */
+#define TRACE_ME		(0x02000000) /* Can be debugged */
+#define TRACE_CALL		(0x04000000) /* Trace calls */
+#define TRACE_REDO		(0x08000000) /* Trace redo */
+#define TRACE_EXIT		(0x10000000) /* Trace edit */
+#define TRACE_FAIL		(0x20000000) /* Trace fail */
+#define FILE_ASSIGNED		(0x40000000) /* Is assigned to a file */
+#define P_REDEFINED		(0x80000000) /* Overrules a definition */
+#define PROC_DEFINED		(P_DYNAMIC|P_FOREIGN|P_MULTIFILE|P_DISCONTIGUOUS)
+
+/* Flags on clauses (packed in unsigned flags : 8) */
+
+#define CL_ERASED		(0x0001) /* clause was erased */
 #define UNIT_CLAUSE		(0x0002) /* Clause has no body */
 #define HAS_BREAKPOINTS		(0x0004) /* Clause has breakpoints */
 #define GOAL_CLAUSE		(0x0008) /* Dummy for meta-calling */
@@ -832,22 +839,32 @@ with one operation, it turns out to be faster as well.
 #define DBREF_CLAUSE		(0x0020) /* Clause has db-reference */
 #define DBREF_ERASED_CLAUSE	(0x0040) /* Deleted while referenced */
 
-#define CHARESCAPE		(0x0004) /* module */
-#define DBLQ_CHARS		(0x0008) /* "ab" --> ['a', 'b'] */
-#define DBLQ_ATOM		(0x0010) /* "ab" --> 'ab' */
-#define DBLQ_STRING		(0x0020) /* "ab" --> "ab" */
+/* Flags on module.  Most of these flags are copied to the read context
+   in pl-read.c.
+*/
+
+#define M_SYSTEM		(0x0001) /* system module */
+#define M_CHARESCAPE		(0x0002) /* module */
+#define DBLQ_CHARS		(0x0004) /* "ab" --> ['a', 'b'] */
+#define DBLQ_ATOM		(0x0008) /* "ab" --> 'ab' */
+#define DBLQ_STRING		(0x0010) /* "ab" --> "ab" */
 #define DBLQ_MASK		(DBLQ_CHARS|DBLQ_ATOM|DBLQ_STRING)
-#define MODULE_COPY_FLAGS	(DBLQ_MASK|CHARESCAPE)
-#define UNKNOWN_FAIL		(0x0040) /* module */
-#define UNKNOWN_WARNING		(0x0080) /* module */
-#define UNKNOWN_ERROR		(0x0100) /* module */
+#define UNKNOWN_FAIL		(0x0020) /* module */
+#define UNKNOWN_WARNING		(0x0040) /* module */
+#define UNKNOWN_ERROR		(0x0080) /* module */
 #define UNKNOWN_MASK		(UNKNOWN_ERROR|UNKNOWN_WARNING|UNKNOWN_FAIL)
 
+/* Flags on functors */
+
 #define CONTROL_F		(0x0002) /* functor (compiled controlstruct) */
 #define ARITH_F			(0x0004) /* functor (arithmetic operator) */
 
+/* Flags on record lists (recorded database keys) */
+
 #define RL_DIRTY		(0x0001) /* recordlist */
 
+/* Flags on recorded database records (also PL_record()) */
+
 #define R_ERASED		(0x0001) /* record: record is erased */
 #define R_EXTERNAL		(0x0002) /* record: inline atoms */
 #define R_DUPLICATE		(0x0004) /* record: include references */
@@ -855,10 +872,9 @@ with one operation, it turns out to be faster as well.
 #define R_DBREF			(0x0010) /* record: has DB-reference */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Handling environment (or local stack) frames.
+Macros for environment frames (local stack frames)
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define FR_BITS			8	/* mask-bits */
 #define FR_HIDE_CHILDS		(0x01L)	/* flag of pred after I_DEPART */
 #define FR_SKIPPED		(0x02L)	/* We have skipped on this frame */
 #define FR_MARKED		(0x04L)	/* GC */
@@ -868,21 +884,11 @@ Handling environment (or local stack) frames.
 #define FR_INBOX		(0x40L) /* Inside box (for REDO in built-in) */
 #define FR_CONTEXT		(0x80L)	/* fr->context is set */
 
-/* FR_LEVEL now handles levels upto 16M.  This is a bit low, but as it is
-   only used for the debugger (skip, etc) it is most likely acceptable.
-   We must consider using a seperate slot in the localFrame
-*/
-
-#define FR_LEVEL		(((uintptr_t)1<<FR_BITS)-1)
-
 #define ARGOFFSET		((int)sizeof(struct localFrame))
 #define VAROFFSET(var)		((var)+(ARGOFFSET/(int)sizeof(word)))
 
-#define setLevelFrame(fr, l)	{ (fr)->flags &= ~FR_LEVEL;   \
-				  (fr)->flags |= ((l) << FR_BITS); \
-				}
-#define levelFrame(fr)		(fr->flags >> FR_BITS)
-#define FR_LEVEL_STEP		((1<<FR_BITS))
+#define setLevelFrame(fr, l)	do { (fr)->level = (l); } while(0)
+#define levelFrame(fr)		((fr)->level)
 #define argFrameP(f, n)		((Word)((f)+1) + (n))
 #define argFrame(f, n)		(*argFrameP((f), (n)) )
 #define varFrameP(f, n)		((Word)(f) + (n))
@@ -890,18 +896,22 @@ Handling environment (or local stack) frames.
 #define refFliP(f, n)		((Word)((f)+1) + (n))
 #define parentFrame(f)		((f)->parent ? (f)->parent\
 					     : (LocalFrame)varFrame((f), -1))
-#define slotsFrame(f)		(true((f)->predicate, FOREIGN) ? \
+#define slotsFrame(f)		(true((f)->predicate, P_FOREIGN) ? \
 				      (f)->predicate->functor->arity : \
 				      (f)->clause->clause->prolog_vars)
 #ifdef O_LOGICAL_UPDATE
+typedef uint64_t gen_t;
 #define generationFrame(f)	((f)->generation)
 #else
 #define generationFrame(f)	(0)
 #endif
 
+#define FR_CLEAR_NEXT	FR_SKIPPED|FR_WATCHED|FR_CATCHED|FR_HIDE_CHILDS
 #define setNextFrameFlags(next, fr) \
-        (next)->flags = ((fr)->flags + FR_LEVEL_STEP) & \
-                        (~(FR_CONTEXT|FR_SKIPPED|FR_WATCHED|FR_CATCHED|FR_HIDE_CHILDS))
+	do \
+	{ (next)->level = (fr)->level+1; \
+	  (next)->flags = ((fr)->flags) & ~(FR_CLEAR_NEXT|FR_CONTEXT); \
+	} while(0)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Predicate reference counting. The aim  of   this  mechanism  is to avoid
@@ -912,13 +922,13 @@ introduce a garbage collector (TBD).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define enterDefinition(def) \
-	if ( unlikely(true(def, DYNAMIC)) ) \
+	if ( unlikely(true(def, P_DYNAMIC)) ) \
 	{ LOCKDYNDEF(def); \
 	  def->references++; \
 	  UNLOCKDYNDEF(def); \
 	}
 #define leaveDefinition(def) \
-	if ( unlikely(true(def, DYNAMIC)) ) \
+	if ( unlikely(true(def, P_DYNAMIC)) ) \
 	{ LOCKDYNDEF(def); \
 	  if ( --def->references == 0 && \
 	       true(def, NEEDSCLAUSEGC) ) \
@@ -1055,18 +1065,13 @@ typedef struct functor_array
 	((cl)->generation.created <= (gen) && \
 	 (cl)->generation.erased   > (gen))
 #else
-#define visibleClause(cl, gen) false(cl, ERASED)
+#define visibleClause(cl, gen) false(cl, CL_ERASED)
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Because struct clause must be a multiple of sizeof(word) for compilation
-on behalf of I_USERCALL the  number  of   shorts  should  be  even. When
-compiling for the stack-shifter we use shorts for the marks slot and the
-line-number, otherwise we use  an  int   for  the  line-number. See also
-WORD_ALIGNED at the declaration of `code'. Demanding word-alignment is a
-machine independent way to achieve   proper alignment, but unfortunately
-it does not port to other C compilers.   Hence the trick with data sizes
-to avoid problems for most platforms.
+Struct clause must be a  multiple   of  sizeof(word)  for compilation on
+behalf  of  I_USERCALL.  This   is   verified    in   an   assertion  in
+checkCodeTable().
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define sizeofClause(n) ((char *)&((Clause)NULL)->codes[n] - (char *)NULL)
@@ -1075,23 +1080,18 @@ struct clause
 { Procedure	procedure;		/* procedure we belong to */
 #ifdef O_LOGICAL_UPDATE
   struct
-  { uintptr_t created;		/* Generation that created me */
-    uintptr_t erased;		/* Generation I was erased */
+  { gen_t created;			/* Generation that created me */
+    gen_t erased;			/* Generation I was erased */
   } generation;
 #endif /*O_LOGICAL_UPDATE*/
   unsigned int		variables;	/* # of variables for frame */
   unsigned int		prolog_vars;	/* # real Prolog variables */
-  unsigned int		line_no;	/* Source line-number */
+  unsigned		flags : 8;	/* Flag field holding: */
+  unsigned		line_no : 24;	/* Source line-number */
   unsigned short	source_no;	/* Index of source-file */
-  unsigned short	flags;		/* Flag field holding: */
-		/* ERASED	   Clause is retracted, but referenced */
-		/* UNIT_CLAUSE     Clause has no body */
-		/* HAS_BREAKPOINTS Break-instructions in the clause */
-		/* GOAL_CLAUSE	   Temporary 'islocal' clause (no head) */
-		/* COMMIT_CLAUSE   Clause will commit (execute !) */
-		/* DBREF_CLAUSE    Clause has a db-reference */
-  code		code_size;		/* size of ->codes */
-  code		codes[1];		/* VM codes of clause */
+  unsigned short	owner_no;	/* Index of owning source-file */
+  code			code_size;	/* size of ->codes */
+  code			codes[1];	/* VM codes of clause */
 };
 
 typedef struct clause_list
@@ -1182,6 +1182,7 @@ struct clause_index
   unsigned	 is_list : 1;		/* Index with lists */
   unsigned int	 dirty;			/* # chains that are dirty */
   float		 speedup;		/* Estimated speedup */
+  struct bit_vector *tried_better;	/* We tried to access for better hash */
   ClauseIndex	 next;			/* Next index */
   ClauseBucket	 entries;		/* chains holding the clauses */
 };
@@ -1198,6 +1199,12 @@ typedef struct local_definitions
   Definition preallocated[7];
 } local_definitions;
 
+#ifdef _MSC_VER
+typedef __int64 meta_mask;		/* MSVC cannot do typedef of typedef!? */
+#else
+typedef uint64_t meta_mask;
+#endif
+
 struct definition
 { FunctorDef	functor;		/* Name/Arity of procedure */
   Module	module;			/* module of the predicate */
@@ -1213,31 +1220,10 @@ struct definition
 #endif
   ClauseIndexList old_clause_indexes;	/* Outdated hash indexes */
   struct bit_vector *tried_index;	/* Arguments on which we tried to index */
+  meta_mask	meta_info;		/* meta-predicate info */
   int		references;		/* reference count */
-  unsigned int  meta_info;		/* meta-predicate info */
   unsigned int  flags;			/* booleans: */
-		/*	FOREIGN		   foreign predicate? */
-		/*	PROFILE_TICKED	   has been ticked this time? */
-		/*	TRACE_ME	   is my call visible? */
-		/*	HIDE_CHILDS	   hide childs for the debugger? */
-		/*	SPY_ME		   spy point set? */
-		/*	DYNAMIC		   dynamic predicate? */
-		/*	MULTIFILE	   defined over more files? */
-		/*	SYSTEM		   system predicate */
-		/*	P_TRANSPARENT	   procedure transparent to modules */
-		/*	DISCONTIGUOUS	   procedure might be discontiguous */
-		/*	NONDETERMINISTIC   deterministic foreign (not used) */
-		/*	GC_SAFE		   Save to perform GC while active */
-		/*	TRACE_CALL	   Trace call-port */
-		/*	TRACE_REDO	   Trace redo-port */
-		/*	TRACE_EXIT	   Trace exit-port */
-		/*	TRACE_FAIL	   Trace fail-port */
-		/*	VOLATILE	   Don't save my clauses */
-		/*	AUTOINDEX	   Automatically guess index */
-		/*	NEEDSCLAUSEGC	   Clauses have been erased */
-		/*	P_VARARG	   Foreign called using t0, ac, ctx */
-		/*	P_SHARED	   Multiple procs are using me */
-  unsigned short shared;		/* #procedures sharing this def */
+  unsigned int  shared;			/* #procedures sharing this def */
 #ifdef O_PROF_PENTIUM
   int		prof_index;		/* index in profiling */
   char	       *prof_name;		/* name in profiling */
@@ -1249,9 +1235,11 @@ struct definition_chain
   DefinitionChain	next;		/* next in chain */
 };
 
+#define	PROC_WEAK	(0x0001)	/* implicit import */
+
 struct procedure
-{ Definition	definition;	/* definition of procedure */
-  int		type;		/* PROCEDURE_TYPE */
+{ Definition	definition;		/* definition of procedure */
+  unsigned int	flags;			/* PROC_WEAK */
 };
 
 struct localFrame
@@ -1264,15 +1252,10 @@ struct localFrame
   struct call_node *prof_node;		/* Profiling node */
 #endif
 #ifdef O_LOGICAL_UPDATE
-  uintptr_t	generation;		/* generation of the database */
+  gen_t		generation;		/* generation of the database */
 #endif
-  unsigned long	flags;			/* packed long holding: */
-		/*	LEVEL	   recursion level (28 bits) */
-		/*	FR_HIDE_CHILDS don't debug this frame ? */
-		/*	FR_SKIPPED skipped in the tracer */
-		/*	FR_MARKED  Marked by GC */
-		/*	FR_WATCHED Watched by the debugger */
-		/*	FR_CATCHED Catched exception here */
+  unsigned int	level;			/* recursion level */
+  unsigned int	flags;			/* packed long holding: */
 };
 
 
@@ -1368,6 +1351,7 @@ struct queryFrame
   int		solutions;		/* # of solutions produced */
   Word	       *aSave;			/* saved argument-stack */
   Choice	saved_bfr;		/* Saved choice-point */
+  LocalFrame	saved_ltop;		/* Saved lTop */
   QueryFrame	parent;			/* Parent queryFrame */
   struct choice	choice;			/* First (dummy) choice-point */
   LocalFrame	saved_environment;	/* Parent local-frame */
@@ -1410,13 +1394,12 @@ struct record
 };
 
 struct recordList
-{ int		type;			/* RECORD_TYPE */
-  int		references;		/* choicepoints reference count */
-  word		key;			/* key of record */
-  RecordRef	firstRecord;		/* first record associated with key */
+{ RecordRef	firstRecord;		/* first record associated with key */
   RecordRef	lastRecord;		/* last record associated with key */
   struct recordList *next;		/* Next recordList */
-  unsigned int  flags;			/* RL_DIRTY */
+  word		key;			/* key of record */
+  unsigned int	flags;			/* RL_DIRTY */
+  int		references;		/* choicepoints reference count */
 };
 
 struct recordRef
@@ -1427,7 +1410,7 @@ struct recordRef
 
 struct sourceFile
 { atom_t	name;			/* name of source file */
-  time_t	time;			/* load time of file */
+  double	mtime;			/* modification time when loaded */
   ListCell	procedures;		/* List of associated procedures */
   Procedure	current_procedure;	/* currently loading one */
   ListCell	modules;		/* Modules associated to this file */
@@ -1460,13 +1443,6 @@ struct module
   int		level;		/* Distance to root (root=0) */
   unsigned int	line_no;	/* Source line-number */
   unsigned int  flags;		/* booleans: */
-		/*	SYSTEM	   system module */
-		/*	DBLQ_INHERIT inherit from default module */
-		/*	DBLQ_CHARS "ab" --> ['a', 'b'] */
-		/*	DBLQ_ATOM  "ab" --> 'ab' */
-		/*	UNKNOWN_FAIL silent failure of unknown pred */
-		/*	UNKNOWN_WARNING Warn on unknown pred */
-		/*	UNKNOWN_ERROR Error on unknown pred */
 };
 
 struct trail_entry
@@ -1487,40 +1463,13 @@ struct gc_trail_entry
 #define MA_ANY		12		/* ? */
 #define MA_NONVAR	13		/* + */
 #define MA_HAT		14		/* ^ */
+#define MA_DCG		15		/* // */
 
-#define MA_INFO(def, n)		(((def)->meta_info >> ((n)*4)) & 0xf)
-#define MA_SETINFO(def, n, i)	((def)->meta_info &= ~(0xf << (n)*4), \
-				 (def)->meta_info |= (i << (n)*4))
-
-
-		 /*******************************
-		 *	 MEMORY ALLOCATION	*
-		 *******************************/
-
-#define ALLOCFAST	(64*SIZEOF_VOIDP) /* big enough for all structures */
-#define ALLOCSIZE	(ALLOCFAST*128)	  /* size of allocation chunks (64K) */
-
-typedef struct free_chunk *FreeChunk;	/* left-over chunk */
-typedef struct chunk *Chunk;		/* Allocation-chunk */
-typedef struct alloc_pool *AllocPool;	/* Allocation pool */
-
-struct chunk
-{ Chunk		next;			/* next of chain */
-};
-
-struct free_chunk
-{ FreeChunk	next;			/* next of chain */
-  size_t	size;			/* size of free bit */
-};
-
-struct alloc_pool
-{ char	       *space;			/* pointer to free space */
-  size_t	free;			/* size of free space */
-  size_t	allocated;		/* total bytes allocated */
-					/* fast perfect fit chains */
-  Chunk		free_chains[ALLOCFAST/sizeof(Chunk)+1];
-  int		free_count[ALLOCFAST/sizeof(Chunk)+1];
-};
+#define MA_INFO(def, n) \
+	(((def)->meta_info >> ((n)*4)) & 0xf)
+#define MA_SETINFO(def, n, i) \
+	((def)->meta_info &= ~((meta_mask)0xf << (n)*4), \
+	 (def)->meta_info |= ((meta_mask)(i) << (n)*4))
 
 
 		 /*******************************
@@ -2052,6 +2001,7 @@ typedef struct debuginfo
 #define PLFLAG_SIGNALS		    0x040000 /* Handle signals */
 #define PLFLAG_DEBUGINFO	    0x080000 /* generate debug info */
 #define PLFLAG_FILEERRORS	    0x100000 /* Edinburgh file errors */
+#define PLFLAG_WARN_OVERRIDE_IMPLICIT_IMPORT 0x200000 /* Warn overriding weak symbols */
 
 typedef struct
 { unsigned int flags;		/* Fast access to some boolean Prolog flags */
@@ -2128,7 +2078,7 @@ decrease).
 #include "pl-funct.ih"
 
 #include "pl-alloc.h"			/* Allocation primitives */
-#include "pl-main.h"			/* Declarations needed by pl-main.c */
+#include "pl-init.h"			/* Declarations needed by pl-init.c */
 #include "pl-error.h"			/* Exception generation */
 #include "pl-thread.h"			/* thread manipulation */
 #include "pl-data.h"			/* Access Prolog data */
@@ -2146,6 +2096,10 @@ decrease).
 #include "os/pl-files.h"		/* File management */
 #include "os/pl-string.h"		/* Basic string functions */
 
+#ifdef ATOMIC_INC
+#define ATOMIC_REFERENCES 1		/* Use atomic +/- for atom references */
+#endif
+
 #ifdef __DECC				/* Dec C-compiler: avoid conflicts */
 #undef leave
 #undef except
@@ -2154,7 +2108,6 @@ decrease).
 
 #ifdef DMALLOC
 #define DMALLOC_FUNC_CHECK 1
-#define O_MYALLOC 0
 #include <dmalloc.h>
 #define allocHeap(n)		malloc(n)
 #define allocHeapOrHalt(n)	xmalloc(n)
