@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -58,8 +57,8 @@ lookupProcedure(functor_t f, Module m)
   } else
   { proc = (Procedure)  allocHeapOrHalt(sizeof(struct procedure));
     def  = (Definition) allocHeapOrHalt(sizeof(struct definition));
-    proc->type = PROCEDURE_TYPE;
     proc->definition = def;
+    proc->flags = 0;
 
     memset(def, 0, sizeof(*def));
     def->functor = valueFunctor(f);
@@ -97,7 +96,7 @@ unallocClauseList(ClauseRef cref)
 
 static void
 unallocDefinition(Definition def)
-{ if ( false(def, FOREIGN|P_THREAD_LOCAL) )
+{ if ( false(def, P_FOREIGN|P_THREAD_LOCAL) )
     unallocClauseList(def->impl.clauses.first_clause);
   else if ( true(def, P_THREAD_LOCAL) )
     free_ldef_vector(def->impl.local);
@@ -106,7 +105,7 @@ unallocDefinition(Definition def)
     freeSimpleMutex(def->mutex);
 
   unallocClauseIndexes(def);
-  freeCodesDefinition(def);
+  freeCodesDefinition(def, FALSE);
 
   freeHeap(def, sizeof(*def));
 }
@@ -124,12 +123,12 @@ unallocProcedure(Procedure proc)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Add (import) a definition to a  module.   Used  by loadImport(). Must be
-merged with pl_import().
+Add (import) a definition to a module.  Used by loadImport() for loading
+states and QLF files. Must be merged with import/1.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-importDefinitionModule(Module m, Definition def)
+importDefinitionModule(Module m, Definition def, int flags)
 { functor_t functor = def->functor->functor;
   Procedure proc;
   int rc = TRUE;
@@ -139,29 +138,28 @@ importDefinitionModule(Module m, Definition def)
   if ( (s = lookupHTable(m->procedures, (void *)functor)) )
   { proc = s->value;
 
-    if ( proc->definition == def )
-      goto done;
-    if ( !isDefinedProcedure(proc) )
-    { Definition odef = proc->definition;
+    if ( proc->definition != def )
+    { if ( !isDefinedProcedure(proc) )
+      { Definition odef = proc->definition;
 
-      shareDefinition(def);
-      proc->definition = def;
-      unshareDefinition(odef);
-      GC_LINGER(odef);
-      goto done;
+	shareDefinition(def);
+	proc->definition = def;
+	unshareDefinition(odef);
+	GC_LINGER(odef);
+      } else
+      { if ( !(flags&PROC_WEAK) )
+	  rc = warning("Failed to import %s into %s",
+		       predicateName(def), PL_atom_chars(m->name));
+      }
     }
-    rc = warning("Failed to import %s into %s",
-		 predicateName(def), PL_atom_chars(m->name));
-    goto done;
   } else
   { proc = (Procedure) allocHeapOrHalt(sizeof(struct procedure));
-    proc->type = PROCEDURE_TYPE;
     proc->definition = def;
+    proc->flags = flags;
     addHTable(m->procedures, (void *)functor, proc);
     shareDefinition(def);
   }
 
-done:
   UNLOCKMODULE(m);
 
   return rc;
@@ -186,7 +184,7 @@ static void
 resetProcedure(Procedure proc, bool isnew)
 { Definition def = proc->definition;
 
-  if ( (true(def, DYNAMIC) && def->references == 0) ||
+  if ( (true(def, P_DYNAMIC) && def->references == 0) ||
        !def->impl.any )
     isnew = TRUE;
 
@@ -198,8 +196,6 @@ resetProcedure(Procedure proc, bool isnew)
   if ( isnew )
   { ClauseIndex ci;
 
-    set(def, AUTOINDEX);
-
     if ( (ci=def->impl.clauses.clause_indexes) )
     { ClauseIndex next;
 
@@ -210,9 +206,9 @@ resetProcedure(Procedure proc, bool isnew)
 	def->impl.clauses.clause_indexes = NULL;
       }
     }
-  }
-
-  freeCodesDefinition(def);		/* carefully sets to S_VIRGIN */
+    freeCodesDefinition(def, FALSE);
+  } else
+    freeCodesDefinition(def, TRUE);	/* carefully sets to S_VIRGIN */
 }
 
 
@@ -231,27 +227,30 @@ ClauseRef
 hasClausesDefinition(Definition def)
 { if ( def->impl.clauses.first_clause )
   { if ( def->impl.clauses.erased_clauses == 0 )
-      return def->impl.clauses.first_clause;
-    else
+    { return def->impl.clauses.first_clause;
+    } else
     { GET_LD
       ClauseRef c;
-      uintptr_t generation;
+      gen_t generation;
       LocalFrame fr = environment_frame;
       if ( fr )
 	generation = generationFrame(fr);
       else
-	generation = ~0L-1;		/* any non-erased clause */
+	generation = (~(gen_t)0)-1;		/* any non-erased clause */
 
+      LOCK();				/* Avoid race with unloadFile() */
       for(c = def->impl.clauses.first_clause; c; c = c->next)
       { Clause cl = c->value.clause;
 
 	if ( visibleClause(cl, generation) )
-	  return c;
+	  break;
       }
+      UNLOCK();
+      return c;
     }
   }
 
-  fail;
+  return NULL;
 }
 
 
@@ -279,15 +278,15 @@ isStaticSystemProcedure(functor_t fd)
   if ( !SYSTEM_MODE &&
        MODULE_system &&
        (proc=isCurrentProcedure(fd, MODULE_system)) &&
-       true(proc->definition, LOCKED) &&
-       false(proc->definition, DYNAMIC) )
+       true(proc->definition, P_LOCKED) &&
+       false(proc->definition, P_DYNAMIC) )
     return proc;
 
   return NULL;
 }
 
 
-static int
+int
 checkModifySystemProc(functor_t fd)
 { Procedure proc;
 
@@ -299,17 +298,64 @@ checkModifySystemProc(functor_t fd)
 }
 
 
+int
+overruleImportedProcedure(Procedure proc, Module target)
+{ GET_LD
+  Definition def = getProcDefinition(proc);
+
+  assert(def->module != target);	/* e.g., imported */
+  if ( true(def->module, M_SYSTEM) )
+  { return PL_error(NULL, 0, NULL, ERR_PERMISSION_PROC,
+		    ATOM_redefine, ATOM_built_in_procedure, proc);
+  } else
+  { if ( proc->flags & PROC_WEAK )
+    { if ( truePrologFlag(PLFLAG_WARN_OVERRIDE_IMPLICIT_IMPORT) )
+      { term_t pi = PL_new_term_ref();
+
+	if ( PL_unify_predicate(pi, proc, GP_NAMEARITY) )
+	{ printMessage(ATOM_warning,
+		       PL_FUNCTOR_CHARS, "ignored_weak_import", 2,
+		         PL_ATOM, target->name,
+		         PL_TERM, pi);
+	}				/* no space to print message */
+      }
+
+      abolishProcedure(proc, target);
+      return TRUE;
+    }
+  }
+
+  return PL_error(NULL, 0, NULL, ERR_PERMISSION_PROC,
+		  ATOM_redefine, ATOM_imported_procedure, proc);
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+lookupProcedureToDefine() locates the proc for  a   functor  in a module
+with the aim of providing a  definition   for  this  procedure, e.g., to
+declare it as a meta-predicate, dynamic, etc.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 Procedure
 lookupProcedureToDefine(functor_t def, Module m)
 { Procedure proc;
 
-  if ( (proc = isCurrentProcedure(def, m)) && false(proc->definition, SYSTEM) )
+  if ( (proc = isCurrentProcedure(def, m)) )
+  { GET_LD
+    Definition def = getProcDefinition(proc);
+
+    if ( def->module != m )
+    { if ( !overruleImportedProcedure(proc, m) )
+	return NULL;
+    }
+
     return proc;
+  }
 
   if ( checkModifySystemProc(def) )
     return lookupProcedure(def, m);
 
-  fail;
+  return NULL;
 }
 
 
@@ -775,7 +821,9 @@ pl_current_predicate1(term_t spec, control_t ctx)
       { Module m = sm->value;
 
 					/* skip hidden modules */
-	if ( stringAtom(m->name)[0] != '$' || SYSTEM_MODE )
+	if ( SYSTEM_MODE ||
+	     m->name == ATOM_system ||
+	     m->class != ATOM_system )
 	  break;
       }
       if ( sm )
@@ -879,11 +927,11 @@ assertProcedure(Procedure proc, Clause clause, int where ARG_LD)
   PL_LOCK(L_MISC);
   clause->generation.created = ++GD->generation;
   PL_UNLOCK(L_MISC);
-  clause->generation.erased  = ~0L;	/* infinite */
+  clause->generation.erased  = ~(gen_t)0;	/* infinite */
 #endif
 
-  if ( false(def, DYNAMIC) )		/* see (*) above */
-    freeCodesDefinition(def);
+  if ( false(def, P_DYNAMIC) )		/* see (*) above */
+    freeCodesDefinition(def, TRUE);
 
   addClauseToIndexes(def, clause, where);
 
@@ -918,7 +966,7 @@ abolishProcedure(Procedure proc, Module module)
     ndef->functor            = def->functor; /* should be merged with */
     ndef->module             = module;	     /* lookupProcedure()!! */
     resetProcedure(proc, TRUE);
-  } else if ( true(def, FOREIGN) )	/* foreign: make normal */
+  } else if ( true(def, P_FOREIGN) )	/* foreign: make normal */
   { def->impl.clauses.first_clause = def->impl.clauses.last_clause = NULL;
     resetProcedure(proc, TRUE);
   } else if ( true(def, P_THREAD_LOCAL) )
@@ -930,7 +978,7 @@ abolishProcedure(Procedure proc, Module module)
   } else				/* normal Prolog procedure */
   { removeClausesProcedure(proc, 0, FALSE);
 
-    if ( true(def, DYNAMIC) )
+    if ( true(def, P_DYNAMIC) )
     { if ( def->references == 0 )
       { ClauseRef cref;
 
@@ -982,10 +1030,10 @@ removeClausesProcedure(Procedure proc, int sfindex, int fromfile)
   for(c = def->impl.clauses.first_clause; c; c = c->next)
   { Clause cl = c->value.clause;
 
-    if ( (sfindex == 0 || sfindex == cl->source_no) &&
+    if ( (sfindex == 0 || sfindex == cl->owner_no) &&
 	 (!fromfile || cl->line_no > 0) &&
-	 false(cl, ERASED) )
-    { set(cl, ERASED);
+	 false(cl, CL_ERASED) )
+    { set(cl, CL_ERASED);
       deleteActiveClauseFromIndexes(def, cl);
 
       if ( deleted++ == 0 )
@@ -1052,14 +1100,14 @@ retractClauseDefinition(Definition def, Clause clause)
 { int rc;
 
   LOCKDYNDEF(def);
-  assert(true(def, DYNAMIC));
-  if ( true(clause, ERASED) )
+  assert(true(def, P_DYNAMIC));
+  if ( true(clause, CL_ERASED) )
   { UNLOCKDYNDEF(def);
     succeed;
   }
 
   DEBUG(CHK_SECURE, checkDefinition(def));
-  set(clause, ERASED);
+  set(clause, CL_ERASED);
 
   if ( def->references ||
        def->impl.clauses.number_of_clauses > 16 )
@@ -1157,7 +1205,7 @@ cleanDefinition(Definition def, ClauseRef garbage)
 	cref=next)
     { next = cref->next;
 
-      if ( true(cref->value.clause, ERASED) )
+      if ( true(cref->value.clause, CL_ERASED) )
       { if ( !prev )
 	{ def->impl.clauses.first_clause = cref->next;
 	  if ( !cref->next )
@@ -1340,7 +1388,7 @@ meta_declaration(term_t spec)
   Definition def;
   atom_t name;
   int i, arity;
-  int mask = 0;
+  meta_mask mask = 0;
   int transparent = FALSE;
 
   if ( !get_procedure(spec, &proc, head, GP_DEFINE) ||
@@ -1348,8 +1396,12 @@ meta_declaration(term_t spec)
     return FALSE;
 
   if ( arity > (int)sizeof(mask)*2 )
-    return PL_error(NULL, 0, "max arity of meta predicates is 8",
+  { char msg[64];
+
+    Ssprintf(msg, "max arity of meta predicates is %d", (int)sizeof(mask)*2);
+    return PL_error(NULL, 0, msg,
 		    ERR_REPRESENTATION, ATOM_max_arity);
+  }
 
   for(i=0; i<arity; i++)
   { atom_t ma;
@@ -1369,13 +1421,15 @@ meta_declaration(term_t spec)
       mask |= e<<(i*4);
       transparent = TRUE;
     } else if ( PL_get_atom(arg, &ma) )
-    { int m;
+    { meta_mask m;
 
       if      ( ma == ATOM_plus )          m = MA_NONVAR;
       else if ( ma == ATOM_minus )         m = MA_VAR;
       else if ( ma == ATOM_question_mark ) m = MA_ANY;
+      else if ( ma == ATOM_star )	   m = MA_ANY; /* * mapped to ? */
       else if ( ma == ATOM_colon )         m = MA_META, transparent = TRUE;
       else if ( ma == ATOM_hat )           m = MA_HAT,  transparent = TRUE;
+      else if ( ma == ATOM_gdiv )          m = MA_DCG,  transparent = TRUE;
       else goto domain_error;
 
       mask |= m<<(i*4);
@@ -1435,6 +1489,7 @@ unify_meta_argument(term_t head, Definition def, int i)
       case MA_ANY:	a = ATOM_question_mark; break;
       case MA_NONVAR:	a = ATOM_plus; break;
       case MA_HAT:	a = ATOM_hat; break;
+      case MA_DCG:	a = ATOM_gdiv; break;
       default:		a = NULL_ATOM; assert(0);
     }
 
@@ -1470,9 +1525,10 @@ PL_meta_predicate(predicate_t proc, const char *spec_s)
   int i;
   int mask = 0;
   int transparent = FALSE;
+  const unsigned char *s = (const unsigned char*)spec_s;
 
-  for(i=0; i<arity; i++)
-  { int spec_c = spec_s[i]&0xff;
+  for(i=0; i<arity; i++, s++)
+  { int spec_c = *s;
     int spec;
 
     switch(spec_c)
@@ -1491,17 +1547,26 @@ PL_meta_predicate(predicate_t proc, const char *spec_s)
       case '^':
 	spec = MA_HAT;
         break;
+      case '/':
+        if ( s[1] == '/' )
+	{ spec = MA_DCG;
+	  s++;
+	  break;
+	} else
+	{ goto invalid;
+	}
       default:
 	if ( spec_c >= '0' && spec_c <= '9' )
 	{ spec = spec_c - '0';
 	  break;
 	}
-        fatalError("Invalid meta-argument\n");
+      invalid:
+        fatalError("Invalid meta-argument for %s: %s\n", procedureName(proc), spec_s);
 	return FALSE;
     }
 
     mask |= spec<<(i*4);
-    if ( spec < 10 || spec == MA_META || spec == MA_HAT )
+    if ( spec < 10 || spec == MA_META || spec == MA_HAT || spec == MA_DCG )
       transparent = TRUE;
   }
 
@@ -1559,7 +1624,7 @@ pl_garbage_collect_clauses(void)
     { Definition def = c->definition;
 
       assert(true(def, P_DIRTYREG));
-      if ( false(def, DYNAMIC) )
+      if ( false(def, P_DYNAMIC) )
 	assert(def->references == 0);
     }
 
@@ -1577,7 +1642,7 @@ pl_garbage_collect_clauses(void)
 
       next = cell->next;
 
-      if ( false(def, DYNAMIC) )
+      if ( false(def, P_DYNAMIC|P_FOREIGN) )
       { if ( def->references )
 	{ assert(def->references == 1);
 	  def->references = 0;
@@ -1635,7 +1700,7 @@ pl_check_definition(term_t spec)
     return Sdprintf("$check_definition/1: can't find definition");
   def = getProcDefinition(proc);
 
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
     succeed;
 
   for(cref = def->impl.clauses.first_clause; cref; cref = cref->next)
@@ -1644,7 +1709,7 @@ pl_check_definition(term_t spec)
     if ( cref->key == 0 )
       nindexable++;
 
-    if ( false(clause, ERASED) )
+    if ( false(clause, CL_ERASED) )
       nclauses++;
     else
       nerased++;
@@ -1761,8 +1826,6 @@ autoLoader(Definition def)
   fid_t  cid;
   term_t argv;
   qid_t qid;
-  atom_t sfn = source_file_name;	/* needs better solution! */
-  int  sln = source_line_no;
   atom_t answer = ATOM_nil;
 
   if ( !GD->procedures.undefinterc4 )
@@ -1777,6 +1840,7 @@ autoLoader(Definition def)
   PL_put_atom(    argv+1, def->functor->name);
   PL_put_integer( argv+2, def->functor->arity);
 
+  push_input_context(ATOM_autoload);
   LD->autoload_nesting++;
   if ( (qid = PL_open_query(MODULE_system, PL_Q_NODEBUG,
 			    GD->procedures.undefinterc4, argv)) )
@@ -1785,8 +1849,7 @@ autoLoader(Definition def)
     PL_close_query(qid);
   }
   LD->autoload_nesting--;
-  source_file_name = sfn;
-  source_line_no   = sln;
+  pop_input_context();
   PL_discard_foreign_frame(cid);
 
   return answer;
@@ -1931,9 +1994,9 @@ PRED_IMPL("retract", 1, retract,
 
       def = getProcDefinition(proc);
 
-      if ( true(def, FOREIGN) )
+      if ( true(def, P_FOREIGN) )
 	return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
-      if ( false(def, DYNAMIC) )
+      if ( false(def, P_DYNAMIC) )
       { if ( isDefinedProcedure(proc) )
 	  return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
 	setDynamicProcedure(proc, TRUE); /* implicit */
@@ -2061,9 +2124,9 @@ pl_retractall(term_t head)
     succeed;
 
   def = getProcDefinition(proc);
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
     return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
-  if ( false(def, DYNAMIC) )
+  if ( false(def, P_DYNAMIC) )
   { if ( isDefinedProcedure(proc) )
       return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
     if ( !setDynamicProcedure(proc, TRUE) )
@@ -2089,7 +2152,7 @@ pl_retractall(term_t head)
 
   DEBUG(CHK_SECURE, checkDefinition(def));
   if ( allvars )
-  { uintptr_t gen = generationFrame(environment_frame);
+  { gen_t gen = generationFrame(environment_frame);
 
     for(cref = def->impl.clauses.first_clause; cref; cref = cref->next)
     { if ( visibleClause(cref->value.clause, gen) )
@@ -2151,7 +2214,7 @@ do_abolish(Module m, term_t atom, term_t arity)
   if ( !(proc = isCurrentProcedure(f, m)) )
     succeed;
 
-  if ( truePrologFlag(PLFLAG_ISO) && false(proc->definition, DYNAMIC) )
+  if ( truePrologFlag(PLFLAG_ISO) && false(proc->definition, P_DYNAMIC) )
     return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
 
   return abolishProcedure(proc, m);
@@ -2188,30 +2251,44 @@ pl_abolish1(term_t spec)		/* Name/Arity */
 }
 
 
-static uintptr_t
-attribute_mask(atom_t key)
-{
+typedef struct patt_mask
+{ atom_t	key;
+  unsigned int  mask;
+} patt_mask;
+
 #define TRACE_ANY (TRACE_CALL|TRACE_REDO|TRACE_EXIT|TRACE_FAIL)
 
-  if (key == ATOM_dynamic)	 return DYNAMIC;
-  if (key == ATOM_multifile)	 return MULTIFILE;
-  if (key == ATOM_system)	 return SYSTEM;
-  if (key == ATOM_locked)	 return LOCKED;
-  if (key == ATOM_spy)		 return SPY_ME;
-  if (key == ATOM_trace)	 return TRACE_ME;
-  if (key == ATOM_trace_call)	 return TRACE_CALL;
-  if (key == ATOM_trace_redo)	 return TRACE_REDO;
-  if (key == ATOM_trace_exit)	 return TRACE_EXIT;
-  if (key == ATOM_trace_fail)	 return TRACE_FAIL;
-  if (key == ATOM_trace_any)	 return TRACE_ANY;
-  if (key == ATOM_hide_childs)	 return HIDE_CHILDS;
-  if (key == ATOM_transparent)	 return P_TRANSPARENT;
-  if (key == ATOM_discontiguous) return DISCONTIGUOUS;
-  if (key == ATOM_volatile)	 return VOLATILE;
-  if (key == ATOM_thread_local)  return P_THREAD_LOCAL;
-  if (key == ATOM_noprofile)     return P_NOPROFILE;
-  if (key == ATOM_iso)		 return P_ISO;
-  if (key == ATOM_public)	 return P_PUBLIC;
+static const patt_mask patt_masks[] =
+{ { ATOM_dynamic,	P_DYNAMIC },
+  { ATOM_multifile,	P_MULTIFILE },
+  { ATOM_locked,	P_LOCKED },
+  { ATOM_system,	P_LOCKED },		/* compatibility */
+  { ATOM_spy,		SPY_ME },
+  { ATOM_trace,		TRACE_ME },
+  { ATOM_trace_call,	TRACE_CALL },
+  { ATOM_trace_redo,	TRACE_REDO },
+  { ATOM_trace_exit,	TRACE_EXIT },
+  { ATOM_trace_fail,	TRACE_FAIL },
+  { ATOM_trace_any,	TRACE_ANY },
+  { ATOM_hide_childs,	HIDE_CHILDS },
+  { ATOM_transparent,	P_TRANSPARENT },
+  { ATOM_discontiguous,	P_DISCONTIGUOUS },
+  { ATOM_volatile,	P_VOLATILE },
+  { ATOM_thread_local,	P_THREAD_LOCAL },
+  { ATOM_noprofile,	P_NOPROFILE },
+  { ATOM_iso,		P_ISO },
+  { ATOM_public,	P_PUBLIC },
+  { (atom_t)0,		0 }
+};
+
+static unsigned int
+attribute_mask(atom_t key)
+{ const patt_mask *p;
+
+  for(p=patt_masks; p->key; p++)
+  { if ( p->key == key )
+      return p->mask;
+  }
 
   return 0;
 }
@@ -2226,7 +2303,7 @@ pl_get_predicate_attribute(term_t pred,
   functor_t fd;
   atom_t key;
   Module module = (Module) NULL;
-  uintptr_t att;
+  unsigned int att;
   term_t head = PL_new_term_ref();
 
   if ( !PL_strip_module(pred, &module, head) ||
@@ -2260,25 +2337,35 @@ pl_get_predicate_attribute(term_t pred,
       d = 0;
 
     return PL_unify_integer(value, d);
-  } else if ( key == ATOM_line_count )
+  } else if ( key == ATOM_line_count || key == ATOM_file )
   { int line;
+    Clause clause;
 
-    if ( false(def, FOREIGN|P_THREAD_LOCAL) &&
+    if ( false(def, P_FOREIGN|P_THREAD_LOCAL) &&
 	 def->impl.clauses.first_clause &&
-	 (line=def->impl.clauses.first_clause->value.clause->line_no) )
-      return PL_unify_integer(value, line);
-    else
-      fail;
+	 (clause = def->impl.clauses.first_clause->value.clause) &&
+	 (line=clause->line_no) )
+    { if ( key == ATOM_line_count )
+      { return PL_unify_integer(value, line);
+      } else
+      { SourceFile sf = indexToSourceFile(clause->source_no);
+
+	if ( sf )
+	  return PL_unify_atom(value, sf->name);
+      }
+    }
+
+    return FALSE;
   } else if ( key == ATOM_foreign )
-  { return PL_unify_integer(value, true(def, FOREIGN) ? 1 : 0);
+  { return PL_unify_integer(value, true(def, P_FOREIGN) ? 1 : 0);
   } else if ( key == ATOM_references )
   { return PL_unify_integer(value, def->references);
   } else if ( key == ATOM_number_of_clauses )
-  { if ( def->flags & FOREIGN )
+  { if ( def->flags & P_FOREIGN )
       fail;
 
     def = getProcDefinition(proc);
-    if ( def->impl.clauses.number_of_clauses == 0 && false(def, DYNAMIC) )
+    if ( def->impl.clauses.number_of_clauses == 0 && false(def, P_DYNAMIC) )
       fail;
     return PL_unify_integer(value, def->impl.clauses.number_of_clauses);
   } else if ( (att = attribute_mask(key)) )
@@ -2340,8 +2427,8 @@ setDynamicProcedure(Procedure proc, bool isdyn)
 { Definition def = proc->definition;
 
   LOCK();
-  if ( (isdyn && true(def, DYNAMIC)) ||
-       (!isdyn && false(def, DYNAMIC)) )
+  if ( (isdyn && true(def, P_DYNAMIC)) ||
+       (!isdyn && false(def, P_DYNAMIC)) )
   { UNLOCK();
     succeed;
   }
@@ -2373,18 +2460,18 @@ setDynamicProcedure(Procedure proc, bool isdyn)
     }
 
   ok:
-    freeCodesDefinition(def);		/* reset to S_VIRGIN */
-    set(def, DYNAMIC);
+    freeCodesDefinition(def, TRUE);	/* reset to S_VIRGIN */
+    set(def, P_DYNAMIC);
 
     UNLOCKDEF(def);
   } else				/* dynamic --> static */
-  { clear(def, DYNAMIC);
+  { clear(def, P_DYNAMIC);
     if ( def->references )
     { if ( true(def, NEEDSCLAUSEGC) )
 	registerDirtyDefinition(def);
       def->references = 0;
     }
-    freeCodesDefinition(def);		/* reset to S_VIRGIN */
+    freeCodesDefinition(def, TRUE);	/* reset to S_VIRGIN */
 
     detachMutexAndUnlock(def);
   }
@@ -2415,7 +2502,7 @@ set_thread_local_procedure(Procedure proc, bool val)
     { UNLOCKDEF(def);
       return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, proc);
     }
-    set(def, DYNAMIC|VOLATILE|P_THREAD_LOCAL);
+    set(def, P_DYNAMIC|P_VOLATILE|P_THREAD_LOCAL);
 
     def->codes = SUPERVISOR(thread_local);
     def->impl.local = new_ldef_vector();
@@ -2431,9 +2518,9 @@ set_thread_local_procedure(Procedure proc, bool val)
   setDynamicProcedure(proc, val);
 
   if ( val )
-    set(proc->definition, VOLATILE|P_THREAD_LOCAL);
+    set(proc->definition, P_VOLATILE|P_THREAD_LOCAL);
   else
-    clear(proc->definition, VOLATILE|P_THREAD_LOCAL);
+    clear(proc->definition, P_VOLATILE|P_THREAD_LOCAL);
 
   succeed;
 #endif
@@ -2466,7 +2553,7 @@ pl_set_predicate_attribute(term_t pred,
   }
   def = proc->definition;
 
-  if ( att == DYNAMIC )
+  if ( att == P_DYNAMIC )
   { rc = setDynamicProcedure(proc, val);
   } else if ( att == P_THREAD_LOCAL )
   { rc = set_thread_local_procedure(proc, val);
@@ -2490,7 +2577,7 @@ pl_set_predicate_attribute(term_t pred,
     addProcedureSourceFile(lookupSourceFile(source_file_name, TRUE), proc);
 
     if ( SYSTEM_MODE )
-    { set(def, SYSTEM|HIDE_CHILDS);
+    { set(def, P_LOCKED|HIDE_CHILDS);
     } else
     { if ( truePrologFlag(PLFLAG_DEBUGINFO) )
 	clear(def, HIDE_CHILDS);
@@ -2517,12 +2604,16 @@ pl_default_predicate(term_t d1, term_t d2)
 }
 
 
-word
-pl_get_clause_attribute(term_t ref, term_t att, term_t value)
+static
+PRED_IMPL("$get_clause_attribute", 3, get_clause_attribute, 0)
 { GET_LD
   Clause clause;
   atom_t a;
 
+  term_t ref   = A1;
+  term_t att   = A2;
+  term_t value = A3;
+
   if ( !PL_get_clref(ref, &clause) ||
        !PL_get_atom_ex(att, &a) )
     return FALSE;
@@ -2533,6 +2624,11 @@ pl_get_clause_attribute(term_t ref, term_t att, term_t value)
   } else if ( a == ATOM_file )
   { SourceFile sf = indexToSourceFile(clause->source_no);
 
+    if ( sf )
+      return PL_unify_atom(value, sf->name);
+  } else if ( a == ATOM_owner )
+  { SourceFile sf = indexToSourceFile(clause->owner_no);
+
     if ( sf )
       return PL_unify_atom(value, sf->name);
   } else if ( a == ATOM_fact )
@@ -2548,6 +2644,11 @@ pl_get_clause_attribute(term_t ref, term_t att, term_t value)
       erased = ATOM_true;
 
     return PL_unify_atom(value, erased);
+  } else if ( a == ATOM_predicate_indicator )
+  { if ( unify_definition(MODULE_user, value,
+			  clause->procedure->definition, 0,
+			  GP_QUALIFY|GP_NAMEARITY) )
+      return TRUE;
   }
 
   fail;
@@ -2742,28 +2843,62 @@ delModuleSourceFile(SourceFile sf, Module m)
 }
 
 
+static void
+delAllModulesSourceFile__unlocked(SourceFile sf)
+{ ListCell c = sf->modules, n;
+
+  sf->modules = NULL;
+
+  for(; c; c = n)
+  { Module m = c->value;
+
+    n = c->next;
+    if ( m->file == sf )
+    { PL_LOCK(L_MODULE);
+      m->file = NULL;
+      m->line_no = 0;
+      clearHTable(m->public);
+      PL_UNLOCK(L_MODULE);
+    }
+
+    freeHeap(c, sizeof(*c));
+  }
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+redefineProcedure() is called when a procedure   needs to be defined and
+it seems to have a definition. The (*)   case occurs if this is actually
+false. This happens if a file holding   a  running predicate is reloaded
+because the clauses cannot be wiped.
+
+Sf is the `owning' source-file
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 redefineProcedure(Procedure proc, SourceFile sf, unsigned int suppress)
 { GET_LD
   Definition def = proc->definition;
 
-  if ( true(def, FOREIGN) )
-  { abolishProcedure(proc, def->module);
+  if ( true(def, P_FOREIGN) )
+  {			/* first call printMessage() */
+			/* so we can provide info about the old definition */
     printMessage(ATOM_warning,
 		 PL_FUNCTOR_CHARS, "redefined_procedure", 2,
 		   PL_CHARS, "foreign",
 		   _PL_PREDICATE_INDICATOR, proc);
-  }
-
-  if ( false(def, MULTIFILE) )
+			/* ... then abolish */
+    abolishProcedure(proc, def->module);
+  } else if ( false(def, P_MULTIFILE) )
   { ClauseRef first;
 
     def = getProcDefinition__LD(def PASS_LD);
-    first = hasClausesDefinition(def);
+    if ( !(first = hasClausesDefinition(def)) )
+      return TRUE;				/* (*) see above */
 
-    if ( first && first->value.clause->source_no == sf->index )
+    if ( first->value.clause->owner_no == sf->index )
     { if ( ((debugstatus.styleCheck & ~suppress) & DISCONTIGUOUS_STYLE) &&
-	   false(def, DISCONTIGUOUS) )
+	   false(def, P_DISCONTIGUOUS) )
 	printMessage(ATOM_warning,
 		     PL_FUNCTOR_CHARS, "discontiguous", 1,
 		       _PL_PREDICATE_INDICATOR, proc);
@@ -2771,8 +2906,6 @@ redefineProcedure(Procedure proc, SourceFile sf, unsigned int suppress)
     { if ( true(def, P_THREAD_LOCAL) )
 	return PL_error(NULL, 0, NULL, ERR_MODIFY_THREAD_LOCAL_PROC, proc);
 
-      abolishProcedure(proc, def->module);
-
       if ( def->references )
       { printMessage(ATOM_informational,
 		     PL_FUNCTOR_CHARS, "redefined_procedure", 2,
@@ -2784,6 +2917,8 @@ redefineProcedure(Procedure proc, SourceFile sf, unsigned int suppress)
 		       PL_CHARS, "static",
 		       _PL_PREDICATE_INDICATOR, proc);
       }
+			/* again, _after_ the printMessage() */
+      abolishProcedure(proc, def->module);
     }
   }
 
@@ -2819,9 +2954,9 @@ pl_source_file(term_t descr, term_t file, control_t h)
   if ( ForeignControl(h) == FRG_FIRST_CALL )
   { if ( get_procedure(descr, &proc, 0, GP_FIND|GP_TYPE_QUIET) )
     { if ( !proc->definition ||
-	   true(proc->definition, FOREIGN|P_THREAD_LOCAL) ||
+	   true(proc->definition, P_FOREIGN|P_THREAD_LOCAL) ||
 	   !(cref = proc->definition->impl.clauses.first_clause) ||
-	   !(sf = indexToSourceFile(cref->value.clause->source_no)) ||
+	   !(sf = indexToSourceFile(cref->value.clause->owner_no)) ||
 	   sf->count == 0 )
 	fail;
 
@@ -2902,7 +3037,7 @@ PRED_IMPL("$time_source_file", 3, time_source_file, PL_FA_NONDETERMINISTIC)
       continue;
 
     if ( PL_unify_atom(file, f->name) &&
-	 unifyTime(time, f->time) &&
+	 PL_unify_float(time, f->mtime) &&
 	 PL_unify_atom(type, f->system ? ATOM_system : ATOM_user) )
     { PL_close_foreign_frame(fid);
       ForeignRedoInt(index+1);
@@ -2970,24 +3105,29 @@ unloadFile(SourceFile sf)
     Procedure proc = cell->value;
     Definition def = proc->definition;
 
-    DEBUG(2, Sdprintf("removeClausesProcedure(%s), refs = %d\n",
-		      predicateName(def), def->references));
+    DEBUG(MSG_UNLOAD, Sdprintf("removeClausesProcedure(%s), refs = %d\n",
+			       predicateName(def), def->references));
 
     deleted = removeClausesProcedure(proc,
-				     true(def, MULTIFILE) ? sf->index : 0,
+				     true(def, P_MULTIFILE) ? sf->index : 0,
 				     TRUE);
 
+    DEBUG(MSG_UNLOAD,
+	  if ( false(def, P_MULTIFILE) && def->impl.clauses.number_of_clauses )
+	    Sdprintf("%s: %d clauses after unload\n",
+		     predicateName(def), def->impl.clauses.number_of_clauses));
+
     if ( deleted )
     { if ( def->references == 0 )
-      { freeCodesDefinition(def);
+      { freeCodesDefinition(def, FALSE);
 	garbage = cleanDefinition(def, garbage);
-      } else if ( false(def, DYNAMIC) )
+      } else if ( false(def, P_DYNAMIC) )
       { registerDirtyDefinition(def);
-	freeCodesDefinition(def);
+	freeCodesDefinition(def, TRUE);
       }
     }
 
-    if ( false(def, MULTIFILE) )
+    if ( false(def, P_MULTIFILE) )
       clear(def, FILE_ASSIGNED);
   }
 
@@ -2996,7 +3136,7 @@ unloadFile(SourceFile sf)
   { Procedure proc = cell->value;
     Definition def = proc->definition;
 
-    if ( false(def, DYNAMIC) && def->references )
+    if ( false(def, P_DYNAMIC) && def->references )
     { assert(def->references == 1);
       def->references = 0;
       GD->procedures.active_marked--;
@@ -3014,6 +3154,7 @@ unloadFile(SourceFile sf)
 #ifdef O_PLMT
   resumeThreads();
 #endif
+  delAllModulesSourceFile__unlocked(sf);
 
   unblockSignals(&set);
   PL_UNLOCK(L_THREAD);
@@ -3110,7 +3251,7 @@ PRED_IMPL("$start_consult", 2, start_consult, 0)
        PL_get_float_ex(modified, &time) )
   { SourceFile f = lookupSourceFile(name, TRUE);
 
-    f->time = (time_t)time;
+    f->mtime = time;
     startConsult(f);
 
     return TRUE;
@@ -3173,7 +3314,7 @@ PRED_IMPL("copy_predicate_clauses", 2, copy_predicate_clauses, PL_FA_TRANSPARENT
   Procedure from, to;
   Definition def, copy_def;
   ClauseRef cref;
-  uintptr_t generation;
+  gen_t generation;
 
   if ( !get_procedure(A1, &from, 0, GP_NAMEARITY|GP_RESOLVE) )
     fail;
@@ -3182,7 +3323,7 @@ PRED_IMPL("copy_predicate_clauses", 2, copy_predicate_clauses, PL_FA_TRANSPARENT
   def = getProcDefinition(from);
   generation = GD->generation;		/* take a consistent snapshot */
 
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
     return PL_error(NULL, 0, NULL, ERR_PERMISSION_PROC,
 		    ATOM_access, ATOM_private_procedure, from);
 
@@ -3190,9 +3331,9 @@ PRED_IMPL("copy_predicate_clauses", 2, copy_predicate_clauses, PL_FA_TRANSPARENT
     return FALSE;
 
   copy_def = getProcDefinition(to);
-  if ( true(copy_def, FOREIGN) )
+  if ( true(copy_def, P_FOREIGN) )
     return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, to);
-  if ( false(copy_def, DYNAMIC) )
+  if ( false(copy_def, P_DYNAMIC) )
   { if ( isDefinedProcedure(to) )
       return PL_error(NULL, 0, NULL, ERR_MODIFY_STATIC_PROC, to);
     if ( !setDynamicProcedure(to, TRUE) )
@@ -3257,7 +3398,7 @@ PRED_IMPL("$clause_from_source", 3, clause_from_source, 0)
   { Procedure proc = cell->value;
     Definition def = proc->definition;
 
-    if ( def && false(def, FOREIGN) )
+    if ( def && false(def, P_FOREIGN) )
     { ClauseRef cref = def->impl.clauses.first_clause;
 
       for( ; cref; cref = cref->next )
@@ -3290,7 +3431,7 @@ PRED_IMPL("$clause_from_source", 3, clause_from_source, 0)
 static void
 listGenerations(Definition def)
 { GET_LD
-  uintptr_t gen = generationFrame(environment_frame);
+  gen_t gen = generationFrame(environment_frame);
   ClauseRef cref;
   int i;
 
@@ -3306,7 +3447,7 @@ listGenerations(Definition def)
 	     clause, i,
 	     clause->generation.created,
 	     clause->generation.erased,
-	     true(clause, ERASED) ? " erased" : "",
+	     true(clause, CL_ERASED) ? " erased" : "",
 	     visibleClause(clause, gen) ? " v " : " X ",
 	     keyName(cref->key));
   }
@@ -3345,7 +3486,7 @@ listGenerations(Definition def)
 		       clauseNo(def, clause),
 		       clause->generation.created,
 		       clause->generation.erased,
-		       true(clause, ERASED) ? " erased" : "",
+		       true(clause, CL_ERASED) ? " erased" : "",
 		       visibleClause(clause, gen) ? " v" : " X");
 	    }
 	  } else
@@ -3356,7 +3497,7 @@ listGenerations(Definition def)
 		     clauseNo(def, clause),
 		     clause->generation.created,
 		     clause->generation.erased,
-		     true(clause, ERASED) ? " erased" : "",
+		     true(clause, CL_ERASED) ? " erased" : "",
 		     visibleClause(clause, gen) ? " v " : " X ",
 		     keyName(cref->key));
 	  }
@@ -3378,7 +3519,7 @@ checkDefinition(Definition def)
   for(nc=0, cref = def->impl.clauses.first_clause; cref; cref=cref->next)
   { Clause clause = cref->value.clause;
 
-    if ( false(clause, ERASED) )
+    if ( false(clause, CL_ERASED) )
     { if ( cref->key )
 	indexed++;
       nc++;
@@ -3412,7 +3553,7 @@ checkDefinition(Definition def)
 	  unsigned int count = 0;
 
 	  for(cr=cl->first_clause; cr; cr=cr->next)
-	  { if ( true(cr->value.clause, ERASED) )
+	  { if ( true(cr->value.clause, CL_ERASED) )
 	      erased++;
 	    else
 	      count++;
@@ -3424,7 +3565,7 @@ checkDefinition(Definition def)
 	} else
 	{ Clause clause = cref->value.clause;
 
-	  if ( true(clause, ERASED) )
+	  if ( true(clause, CL_ERASED) )
 	    dirty++;
 	}
       }
@@ -3450,7 +3591,7 @@ pl_check_procedure(term_t desc)
     fail;
   def = getProcDefinition(proc);
 
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
     fail;
 
   checkDefinition(def);
@@ -3469,7 +3610,7 @@ pl_list_generations(term_t desc)
     fail;
   def = getProcDefinition(proc);
 
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
     fail;				/* permission error */
 
   listGenerations(def);
@@ -3489,6 +3630,7 @@ BeginPredDefs(proc)
   PRED_DEF("meta_predicate", 1, meta_predicate, PL_FA_TRANSPARENT)
   PRED_DEF("$time_source_file", 3, time_source_file, PL_FA_NONDETERMINISTIC)
   PRED_DEF("$clause_from_source", 3, clause_from_source, 0)
+  PRED_DEF("$get_clause_attribute", 3, get_clause_attribute, 0)
   PRED_DEF("retract", 1, retract,
 	   PL_FA_TRANSPARENT|PL_FA_NONDETERMINISTIC|PL_FA_ISO)
   PRED_DEF("$unload_file", 1, unload_file, 0)
