@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -30,7 +29,7 @@
 #define WORDS_PER_PLINT (sizeof(int64_t)/sizeof(word))
 
 static RecordList lookupRecordList(word);
-static RecordList isCurrentRecordList(word);
+static RecordList isCurrentRecordList(word, int must_be_non_empty);
 static void freeRecordRef(RecordRef r);
 static void unallocRecordList(RecordList rl);
 
@@ -86,7 +85,6 @@ lookupRecordList(word key)
       PL_register_atom(key);
     l = allocHeapOrHalt(sizeof(*l));
     l->key = key;
-    l->type = RECORD_TYPE;
     l->references = 0;
     l->flags = 0;
     l->firstRecord = l->lastRecord = NULL;
@@ -105,11 +103,26 @@ lookupRecordList(word key)
 
 
 static RecordList
-isCurrentRecordList(word key)
+isCurrentRecordList(word key, int must_be_non_empty)
 { Symbol s;
 
   if ( (s = lookupHTable(GD->recorded_db.record_lists, (void *)key)) )
-    return s->value;
+  { RecordList rl = s->value;
+
+    if ( must_be_non_empty )
+    { RecordRef record;
+
+      LOCK();
+      for(record = rl->firstRecord; record; record = record->next)
+      { if ( false(record->record, R_ERASED) )
+	  break;
+      }
+      UNLOCK();
+      return record ? rl : NULL;
+    } else
+    { return rl;
+    }
+  }
 
   return NULL;
 }
@@ -578,8 +591,19 @@ restoreVars(compile_info *info)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+compileTermToHeap__LD() is the core of the recorded database.
+
+Returns NULL if there is insufficient   memory.  Otherwise the result of
+the  allocation  function.   The   default    allocation   function   is
+PL_malloc_atomic_unmanaged().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 Record
-compileTermToHeap__LD(term_t t, int flags ARG_LD)
+compileTermToHeap__LD(term_t t,
+		      void* (*allocate)(void *closure, size_t size),
+		      void* closure,
+		      int flags ARG_LD)
 { compile_info info;
   Record record;
   size_t size;
@@ -603,18 +627,25 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
   unvisit(PASS_LD1);
 
   size = rsize + sizeOfBuffer(&info.code);
-  record = PL_malloc_atomic_unmanaged(size);
+  if ( allocate )
+    record = (*allocate)(closure, size);
+  else
+    record = PL_malloc_atomic_unmanaged(size);
+
+  if ( record )
+  {
 #ifdef REC_MAGIC
-  record->magic = REC_MAGIC;
+    record->magic = REC_MAGIC;
 #endif
-  record->gsize = (unsigned int)info.size; /* only 28-bit */
-  record->nvars = info.nvars;
-  record->size  = (int)size;
-  record->flags = flags;
-  if ( flags & R_DUPLICATE )
-  { record->references = 1;
+    record->gsize = (unsigned int)info.size; /* only 28-bit */
+    record->nvars = info.nvars;
+    record->size  = (int)size;
+    record->flags = flags;
+    if ( flags & R_DUPLICATE )
+    { record->references = 1;
+    }
+    memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
   }
-  memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
   discardBuffer(&info.code);
 
   DEBUG(3, Sdprintf("--> record at %p\n", record));
@@ -714,7 +745,7 @@ PL_record_external(term_t t, size_t *len)
     addUintBuffer((Buffer)&hdr, info.nvars);	/* Number of variables */
   shdr = (int)sizeOfBuffer(&hdr);
 
-  rec = PL_malloc_atomic(shdr + scode);
+  rec = PL_malloc_atomic_unmanaged(shdr + scode);
   memcpy(rec, hdr.base, shdr);
   memcpy(rec+shdr, info.code.base, scode);
 
@@ -1529,7 +1560,6 @@ PL_recorded_external(const char *rec, term_t t)
 int
 PL_erase_external(char *rec)
 { copy_info b;
-  uint scode;
   uchar m;
 
   b.base = b.data = rec;
@@ -1538,20 +1568,8 @@ PL_erase_external(char *rec)
   { Sdprintf("PL_erase_external(): incompatible version\n");
     fail;
   }
-  if (  m & (REC_INT|REC_ATOM) )
-  { if (  m & REC_INT )
-      skipLong(&b);
-    else
-      skipAtom(&b);
-  } else
-  { scode = fetchSizeInt(&b);
-    skipSizeInt(&b);			/* gsize */
-    if ( !(m & REC_GROUND) )
-      skipSizeInt(&b);			/* nvars */
-    b.data += scode;
-  }
 
-  freeHeap(rec, b.data-b.base);
+  PL_free(rec);
   return TRUE;
 }
 
@@ -1605,7 +1623,7 @@ PRED_IMPL("current_key", 1, current_key, PL_FA_NONDETERMINISTIC)
       { rl = GD->recorded_db.head;
 	break;
       } else if ( getKeyEx(A1, &k PASS_LD) &&
-		  isCurrentRecordList(k) )
+		  isCurrentRecordList(k, TRUE) )
 	succeed;
 
       fail;
@@ -1655,7 +1673,8 @@ record(term_t key, term_t term, term_t ref, int az)
   if ( ref && !PL_is_variable(ref) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_variable, ref);
 
-  copy = compileTermToHeap(term, 0);
+  if ( !(copy = compileTermToHeap(term, 0)) )
+    return PL_no_memory();
   r = allocHeapOrHalt(sizeof(*r));
   r->record = copy;
   if ( ref && !PL_unify_recref(ref, r) )
@@ -1739,7 +1758,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 	  fail;
 	varkey = TRUE;
       } else if ( getKeyEx(key, &k PASS_LD) )
-      { if ( !(rl = isCurrentRecordList(k)) )
+      { if ( !(rl = isCurrentRecordList(k, FALSE)) )
 	  fail;
 	varkey = FALSE;
       } else
@@ -1869,7 +1888,7 @@ PRED_IMPL("instance", 2, instance, 0)
 
   if ( type == DB_REF_CLAUSE )
   { Clause clause = ptr;
-    uintptr_t generation = environment_frame->generation;
+    gen_t generation = environment_frame->generation;
 
     if ( true(clause, GOAL_CLAUSE) ||
 	 !visibleClause(clause, generation) )
@@ -1917,7 +1936,7 @@ PRED_IMPL("erase", 1, erase, 0)
   { Clause clause = ptr;
     Definition def = getProcDefinition(clause->procedure);
 
-    if ( !true(def, DYNAMIC) )
+    if ( !true(def, P_DYNAMIC) )
       return PL_error("erase", 1, NULL, ERR_PERMISSION,
 		      ATOM_clause, ATOM_erase, ref);
 
