@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -119,6 +118,7 @@ typedef struct message_queue
   int		       waiting;		/* # waiting threads */
   int		       waiting_var;	/* # waiting with unbound */
   int		       wait_for_drain;	/* # threads waiting for write */
+  unsigned	initialized : 1;	/* Queue is initialised */
   unsigned	destroyed : 1;		/* Thread is being destroyed */
   unsigned	type : 2;		/* QTYPE_* */
 } message_queue;
@@ -215,7 +215,7 @@ compile-time
 	if ( GD->thread.enabled ) \
 	{ if ( def->mutex ) \
 	  { countingMutexLock(def->mutex); \
-	  } else if ( false(def, DYNAMIC) ) \
+	  } else if ( false(def, P_DYNAMIC) ) \
 	  { countingMutexLock(&_PL_mutexes[L_PREDICATE]); \
 	  } \
 	}
@@ -224,7 +224,7 @@ compile-time
 	if ( GD->thread.enabled ) \
 	{ if ( def->mutex ) \
 	  { countingMutexUnlock(def->mutex); \
-	  } else if ( false(def, DYNAMIC) ) \
+	  } else if ( false(def, P_DYNAMIC) ) \
 	  { countingMutexUnlock(&_PL_mutexes[L_PREDICATE]); \
 	  } \
 	}
@@ -312,7 +312,6 @@ COMMON(int)		exitPrologThreads(void);
 COMMON(bool)		aliasThread(int tid, atom_t name);
 COMMON(word)		pl_thread_create(term_t goal, term_t id,
 					 term_t options);
-COMMON(word)		pl_thread_join(term_t thread, term_t retcode);
 COMMON(word)		pl_thread_exit(term_t retcode);
 COMMON(foreign_t)	pl_thread_signal(term_t thread, term_t goal);
 
@@ -328,7 +327,6 @@ COMMON(foreign_t)	pl_mutex_unlock_all(void);
 COMMON(const char *)	threadName(int id);
 COMMON(void)		executeThreadSignals(int sig);
 COMMON(foreign_t)	pl_attach_xterm(term_t in, term_t out);
-COMMON(size_t)		threadLocalHeapUsed(void);
 COMMON(int)		attachConsole(void);
 COMMON(Definition)	localiseDefinition(Definition def);
 COMMON(LocalDefinitions) new_ldef_vector(void);
@@ -341,14 +339,16 @@ COMMON(void)		cleanupThreads();
 COMMON(intptr_t)	system_thread_id(PL_thread_info_t *info);
 COMMON(double)	        ThreadCPUTime(PL_local_data_t *ld, int which);
 
+
 		 /*******************************
 		 *	 GLOBAL GC SUPPORT	*
 		 *******************************/
 
-void		forThreadLocalData(void (*func)(struct PL_local_data *),
+COMMON(void)	forThreadLocalData(void (*func)(struct PL_local_data *),
 				   unsigned flags);
-void		resumeThreads(void);
-void		markAtomsThreads(void);
+COMMON(void)	resumeThreads(void);
+COMMON(void)	markAtomsMessageQueues(void);
+COMMON(void)	markAtomsThreadMessageQueue(PL_local_data_t *ld);
 
 #define PL_THREAD_SUSPEND_AFTER_WORK	0x1 /* forThreadLocalData() */
 
